import os
import numpy as np
from skimage import img_as_float32
from skimage.draw import polygon
import urllib.request
import zipfile
import shutil
import tempfile

import dask.array as da
import napari
from napari.layers import Labels
from napari.utils.colormaps import DirectLabelColormap
from napari.utils.notifications import show_info
from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QLabel, QComboBox, QDoubleSpinBox, QGridLayout, QSizePolicy, QFrame, QApplication, QFileDialog, QMessageBox
from qtpy.QtGui import QPixmap, QIcon
from qtpy.QtCore import Qt

from csbdeep.utils import normalize
from stardist.utils import edt_prob
from stardist.geometry import star_dist
from napari_hydra.hydrastardist.models.model2d_hydra import Config2D, StarDist2D
from napari_hydra.utils import rasterize_labels, ensure_default_model, is_image_stack, process_frame, get_well_centers, sort_wells_grid, calculate_well_stats, calculate_well_diameters, create_hydra_colormaps, get_or_create_layer, prepare_training_batches, write_prediction_summary

import tensorflow as tf

import warnings
warnings.filterwarnings("ignore", category=UserWarning, module="bioimageio_utils")

class HydraStarDistPlugin(QWidget):
    """
    Hydra StarDist Plugin for Napari
    """
    # Default model ZIP to download when no local models are present
    ZIP_URL = "https://rodare.hzdr.de/record/4439/files/model_vacvplaque_hsd.zip?download=1"
    
    # Constants
    WELL_DIAMETER_MM = 35

    def __init__(self, viewer: napari.Viewer = None):
        """
        Initialized the plugin:
            - Loads Hydra StarDist model.
            - Sets up the UI layout.
            - Connects buttons to respective functions.
        """
        super().__init__()
        self.viewer = viewer
        self.setWindowTitle("Hydra StarDist Plugin")
        plugin_dir = os.path.dirname(__file__)
        
        """
        UI Structure:
            - Header
            - Prediction
                - Selection of image, model, and compression size.
                - Finetuning of thresholds parameters.
            - Counting
                - Counts plaques in each well.
            - Tuning
                - Allows to callibrate the model.
        """
        layout = QVBoxLayout()

        ## HEADER
        ## ---------------------------------------------------------------------------------
        # displays logo and small text
        top_layout = QVBoxLayout()
        top_layout.setAlignment(Qt.AlignHCenter)

        # Logo 
        logo_label = QLabel()
        logo_path = os.path.join(plugin_dir, "resources", "hydrastardist_logo.png")  # makes sure logo exists
        pixmap = QPixmap(logo_path)
        pixmap = pixmap.scaledToWidth(250, mode=Qt.SmoothTransformation)
        logo_label.setPixmap(pixmap)
        logo_label.setAlignment(Qt.AlignCenter)
        top_layout.addWidget(logo_label)

        # Small Text
        text_label = QLabel("HydraStarDist is a branched version of StarDist, allowing for star-convex object detection of two different circular targets at the same time.")
        text_label.setWordWrap(True)
        text_label.setAlignment(Qt.AlignJustify)
        text_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        top_layout.addWidget(text_label)

        layout.addLayout(top_layout)
        layout.addSpacing(10)

        ## COMMON ATTRIBUTES
        ## ---------------------------------------------------------------------------------
        # IMAGE AND MODEL
        selection_grid = QGridLayout()
        selection_grid.setHorizontalSpacing(10)
        selection_grid.setVerticalSpacing(5)

        selection_grid.addWidget(QLabel("Image"), 0, 0)
        # Subclass QComboBox to refresh image layers when dropdown is opened
        class RefreshOnShowComboBox(QComboBox):
            def __init__(combo_self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                # Reference to plugin instance for refresh
                combo_self._plugin = self
            def showPopup(combo_self):
                combo_self._plugin.refresh_image_layers()
                super().showPopup()
        self.image_layer_combo = RefreshOnShowComboBox()
        selection_grid.addWidget(self.image_layer_combo, 1, 0)

        selection_grid.addWidget(QLabel("Model"), 0, 1)
        self.model_combo = QComboBox()
        model_basedir = os.path.join(plugin_dir, "resources", "models")

        # Ensure models directory exists
        os.makedirs(model_basedir, exist_ok=True)

        # If no model subfolders are present, attempt to download the default model zip
        self.model_names = ensure_default_model(model_basedir, HydraStarDistPlugin.ZIP_URL)
        self.model_combo.addItems(self.model_names)

        if "VACV-Plaque" in self.model_names:
            idx = self.model_names.index("VACV-Plaque")
            self.model_combo.setCurrentIndex(idx)
        selection_grid.addWidget(self.model_combo, 1, 1)
        layout.addLayout(selection_grid)
        layout.addSpacing(5)

        if not self.model_names:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("Default model download failed.")
            msg.setInformativeText("Please check your internet connection and try again.")
            msg.setWindowTitle("Download Error")
            msg.exec_()
            return

        # LOAD HYDRASTARDIST MODEL
        # searches for model and loads "VACV-Plaque" as default
        self.model = StarDist2D(
            None,
            name="VACV-Plaque",
            basedir=model_basedir
        )

        # connects model combo to change_model slot
        self.model_combo.currentIndexChanged.connect(self.change_model)

        # COMPRESSION SIZE
        target_size_layout = QHBoxLayout()
        target_size_label = QLabel("Resize")
        target_size_layout.addWidget(target_size_label)
        info_icon = QIcon(os.path.join(plugin_dir, "resources", "info.png"))
        info_label = QLabel()
        info_label.setPixmap(info_icon.pixmap(18, 18))
        info_label.setToolTip(
            "Image is compressed for quicker model inference.\n"
            "These are dimensions to which the image will be resized.\n"
            "In case there is a different aspect ratio, the image is\n"
            "cropped and the aspect ratio is kept."
        )
        target_size_layout.addStretch()
        target_size_layout.addWidget(info_label)
        input_size = getattr(self.model.config, "input_size", [800, 600])
        self.target_width_spin = QDoubleSpinBox()
        self.target_width_spin.setRange(1, 4096)
        self.target_width_spin.setDecimals(0)
        self.target_width_spin.setValue(input_size[0])
        self.target_width_spin.setFixedWidth(70)
        self.target_width_spin.setAlignment(Qt.AlignCenter)
        self.target_width_spin.setStyleSheet("padding: 2px;")
        target_size_layout.addWidget(self.target_width_spin)
        target_size_layout.addWidget(QLabel("x"))
        self.target_height_spin = QDoubleSpinBox()
        self.target_height_spin.setRange(1, 4096)
        self.target_height_spin.setDecimals(0)
        self.target_height_spin.setValue(input_size[1])
        self.target_height_spin.setFixedWidth(70)
        self.target_height_spin.setAlignment(Qt.AlignCenter)
        self.target_height_spin.setStyleSheet("padding: 2px;")
        target_size_layout.addWidget(self.target_height_spin)
        layout.addLayout(target_size_layout)
        layout.addSpacing(10)

        ## PREDCITION
        ## ---------------------------------------------------------------------------------
        prediction_header_layout = QHBoxLayout()
        prediction_label = QLabel("<h2>Prediction</h2>")
        prediction_header_layout.addWidget(prediction_label)
        prediction_header_layout.addStretch()
        layout.addLayout(prediction_header_layout)
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)
        separator.setFixedHeight(2)
        separator.setStyleSheet("background-color: gray;")
        layout.addWidget(separator)

        # THRESHOLDING
        threshold_layout = QVBoxLayout()
        reload_icon = QIcon(os.path.join(plugin_dir, "resources", "reload.png"))
        info_icon   = QIcon(os.path.join(plugin_dir, "resources", "info.png"))

        # Table-like grid using QGridLayout
        grid = QGridLayout()
        grid.setHorizontalSpacing(10)
        grid.setVerticalSpacing(5)

        # Threshold Header
        threshold_header_widget = QWidget()
        threshold_header_layout = QHBoxLayout()
        threshold_header_layout.setContentsMargins(0, 0, 0, 0)
        threshold_header_layout.setSpacing(5)
        threshold_label = QLabel("<b>Thresholds</b>")
        threshold_label.setAlignment(Qt.AlignCenter)
        threshold_header_layout.addWidget(threshold_label)

        info_label = QLabel()
        info_label.setPixmap(info_icon.pixmap(18, 18))
        info_label.setToolTip(
            "Thresholds control object detection applied \n"
            "to the HydraStarDist model output."
        )
        threshold_header_layout.addStretch()
        threshold_header_layout.addWidget(info_label)
        threshold_header_widget.setLayout(threshold_header_layout)
        grid.addWidget(threshold_header_widget, 0, 0)

        # Wells Header
        wells_header_widget = QWidget()
        wells_header_layout = QHBoxLayout()
        wells_header_layout.setContentsMargins(0, 0, 0, 0)
        wells_header_layout.setSpacing(5)
        wells_label = QLabel("Wells")
        wells_label.setAlignment(Qt.AlignCenter)
        self.reset_wells_btn = QPushButton()
        self.reset_wells_btn.setIcon(reload_icon)
        self.reset_wells_btn.setToolTip(
            "Reset well thresholds.")
        self.reset_wells_btn.setFixedSize(24, 24)
        self.reset_wells_btn.clicked.connect(self.reset_wells_thresholds)
        wells_header_layout.addWidget(wells_label)
        wells_header_layout.addWidget(self.reset_wells_btn)
        wells_header_widget.setLayout(wells_header_layout)
        grid.addWidget(wells_header_widget, 0, 1)

        # Plaques Header
        plaques_header_widget = QWidget()
        plaques_header_layout = QHBoxLayout()
        plaques_header_layout.setContentsMargins(0, 0, 0, 0)
        plaques_header_layout.setSpacing(5)
        plaques_label = QLabel("Plaques")
        plaques_label.setAlignment(Qt.AlignCenter)
        self.reset_plaques_btn = QPushButton()
        self.reset_plaques_btn.setIcon(reload_icon)
        self.reset_plaques_btn.setToolTip(
            "Reset Plaques thresholds."
        )
        self.reset_plaques_btn.setFixedSize(24, 24)
        self.reset_plaques_btn.clicked.connect(self.reset_plaques_thresholds)
        plaques_header_layout.addWidget(plaques_label)
        plaques_header_layout.addWidget(self.reset_plaques_btn)
        plaques_header_widget.setLayout(plaques_header_layout)
        grid.addWidget(plaques_header_widget, 0, 2)

        # Probability Row
        # Probability Label
        probability_header_widget = QWidget()
        probability_header_layout = QHBoxLayout()
        probability_header_layout.setContentsMargins(0, 0, 0, 0)
        probability_header_layout.setSpacing(5)
        probability_label = QLabel("Probability")
        probability_label.setAlignment(Qt.AlignCenter)
        probability_label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        probability_header_layout.addWidget(probability_label)
        
        info_label = QLabel()
        info_label.setPixmap(info_icon.pixmap(18, 18))
        info_label.setToolTip(
            "Defines how strongly the object must resemble a star-convex object."
        )
        probability_header_layout.addStretch()
        probability_header_layout.addWidget(info_label)
        probability_header_widget.setLayout(probability_header_layout)
        grid.addWidget(probability_header_widget, 1, 0)

        #  Well Probability
        self.wells_prob_spin = QDoubleSpinBox()
        self.wells_prob_spin.setRange(0.0, 1.0)
        self.wells_prob_spin.setSingleStep(0.01)
        self.wells_prob_spin.setValue(self.model.thresholds1['prob'])
        self.wells_prob_spin.setAlignment(Qt.AlignCenter)
        self.wells_prob_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.wells_prob_spin.setStyleSheet("padding: 2px;")
        grid.addWidget(self.wells_prob_spin, 1, 1)

        #  Plaque Probability
        self.plaques_prob_spin = QDoubleSpinBox()
        self.plaques_prob_spin.setRange(0.0, 1.0)
        self.plaques_prob_spin.setSingleStep(0.01)
        self.plaques_prob_spin.setValue(self.model.thresholds2['prob'])
        self.plaques_prob_spin.setAlignment(Qt.AlignCenter)
        self.plaques_prob_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.plaques_prob_spin.setStyleSheet("padding: 2px;")
        grid.addWidget(self.plaques_prob_spin, 1, 2)

        # NMS Row
        # NMS Label
        nms_header_widget = QWidget()
        nms_header_layout = QHBoxLayout()
        nms_header_layout.setContentsMargins(0, 0, 0, 0)
        nms_header_layout.setSpacing(5)
        nms_label = QLabel("NMS")
        nms_label.setAlignment(Qt.AlignCenter)
        nms_label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        nms_header_layout.addWidget(nms_label)

        info_label = QLabel()
        info_label.setPixmap(info_icon.pixmap(18, 18))
        info_label.setToolTip(
            "Defines how strong the non-maximum suppression is."
        )
        nms_header_layout.addStretch()
        nms_header_layout.addWidget(info_label)
        nms_header_widget.setLayout(nms_header_layout)
        grid.addWidget(nms_header_widget, 2, 0)

        # Well NMS
        self.wells_overlap_spin = QDoubleSpinBox()
        self.wells_overlap_spin.setRange(0.0, 1.0)
        self.wells_overlap_spin.setSingleStep(0.01)
        self.wells_overlap_spin.setValue(self.model.thresholds1['nms'])
        self.wells_overlap_spin.setAlignment(Qt.AlignCenter)
        self.wells_overlap_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.wells_overlap_spin.setStyleSheet("padding: 2px;")
        grid.addWidget(self.wells_overlap_spin, 2, 1)

        # Plaque NMS
        self.plaques_overlap_spin = QDoubleSpinBox()
        self.plaques_overlap_spin.setRange(0.0, 1.0)
        self.plaques_overlap_spin.setSingleStep(0.01)
        self.plaques_overlap_spin.setValue(self.model.thresholds2['nms'])
        self.plaques_overlap_spin.setAlignment(Qt.AlignCenter)
        self.plaques_overlap_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.plaques_overlap_spin.setStyleSheet("padding: 2px;")
        grid.addWidget(self.plaques_overlap_spin, 2, 2)

        threshold_layout.addLayout(grid)
        layout.addLayout(threshold_layout)
        
        # Prediction Button
        layout.addSpacing(5)
        # Prediction and Export Buttons (side by side)
        prediction_btn_layout = QHBoxLayout()
        self.predict_btn = QPushButton("Run Prediction")
        self.predict_btn.clicked.connect(self.run_prediction)
        prediction_btn_layout.addWidget(self.predict_btn)
        # Add Export button, smaller and next to Run Prediction
        self.export_btn = QPushButton("Export")
        self.export_btn.setFixedWidth(70)
        self.export_btn.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.export_btn.clicked.connect(self.export_prediction)
        prediction_btn_layout.addWidget(self.export_btn)
        layout.addLayout(prediction_btn_layout)

        ## COUNTING
        ## ---------------------------------------------------------------------------------
        layout.addSpacing(10)
        counting_header_layout = QHBoxLayout()
        counting_label = QLabel("<h2>Counting</h2>")
        counting_header_layout.addWidget(counting_label)
        counting_header_layout.addStretch()
        layout.addLayout(counting_header_layout)

        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)
        separator.setFixedHeight(2)
        separator.setStyleSheet("background-color: gray;")
        layout.addWidget(separator)
        
        # Grid for Well Counts
        grid_counts = QGridLayout()
        self.well_count_labels = []
        for i in range(6):
            label = QLabel("")
            label.setAlignment(Qt.AlignCenter)
            label.setFrameStyle(QFrame.Panel | QFrame.Sunken)
            label.setMinimumWidth(40)
            label.setStyleSheet("background-color: #414851; border-radius: 2px; padding: 2px;")
            self.well_count_labels.append(label)
            grid_counts.addWidget(label, i // 3, i % 3)
        layout.addLayout(grid_counts)

        # Plaque Counts Section
        counts_header_layout = QHBoxLayout()
        counts_label = QLabel("Plaque Count per Well")
        counts_header_layout.addWidget(counts_label)
        counts_header_layout.addStretch()
        self.count_btn = QPushButton("Count")
        self.count_btn.clicked.connect(self.count_plaque)
        counts_header_layout.addWidget(self.count_btn)
        layout.addLayout(counts_header_layout)

        ## TUNING
        ## ---------------------------------------------------------------------------------
        layout.addSpacing(10)
        tuning_header_layout = QHBoxLayout()
        tuning_label = QLabel("<h2>Tuning</h2>")
        tuning_header_layout.addWidget(tuning_label)
        tuning_header_layout.addStretch()
        layout.addLayout(tuning_header_layout)
        
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)
        separator.setFixedHeight(2)
        separator.setStyleSheet("background-color: gray;")
        layout.addWidget(separator)

        ## HYPERPARAMETERS
        tuning_grid = QGridLayout()
        tuning_grid.setHorizontalSpacing(10)
        tuning_grid.setVerticalSpacing(5)

        # Batch Size
        tuning_grid.addWidget(QLabel("Batch Size"), 0, 0)
        self.batch_spin = QDoubleSpinBox()
        self.batch_spin.setRange(1, 1000)
        self.batch_spin.setSingleStep(1)
        self.batch_spin.setDecimals(0)
        self.batch_spin.setValue(5)
        self.batch_spin.setAlignment(Qt.AlignCenter)
        self.batch_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.batch_spin.setStyleSheet("padding: 2px;")
        tuning_grid.addWidget(self.batch_spin, 1, 0)

        # Epochs
        tuning_grid.addWidget(QLabel("Epochs"), 0, 1)
        self.epochs_spin = QDoubleSpinBox()
        self.epochs_spin.setRange(0, 400)
        self.epochs_spin.setSingleStep(1)
        self.epochs_spin.setDecimals(0)
        self.epochs_spin.setValue(10)
        self.epochs_spin.setAlignment(Qt.AlignCenter)
        self.epochs_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.epochs_spin.setStyleSheet("padding: 2px;")
        tuning_grid.addWidget(self.epochs_spin, 1, 1)

        # Learning Rate
        tuning_grid.addWidget(QLabel("Learning Rate"), 0, 2)
        self.lr_spin = QDoubleSpinBox()
        self.lr_spin.setRange(0, 1)
        self.lr_spin.setSingleStep(0.0001)
        self.lr_spin.setDecimals(4)
        self.lr_spin.setValue(0.001)
        self.lr_spin.setAlignment(Qt.AlignCenter)
        self.lr_spin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.lr_spin.setStyleSheet("padding: 2px;")
        tuning_grid.addWidget(self.lr_spin, 1, 2)

        layout.addLayout(tuning_grid)
        layout.addSpacing(5)

        # Tune Button
        self.tune_btn = QPushButton("Tune Model")
        self.tune_btn.clicked.connect(self.tune_model)
        layout.addWidget(self.tune_btn)

        self.setLayout(layout)

        # Fill combo with current image layers
        self.refresh_image_layers()

        # In case labels are uploaded, they are compatible with painting operations
        def ensure_numpy_on_add(event):
            layer = event.value
            if isinstance(layer, Labels) and isinstance(layer.data, da.Array):
                print(f"Converting {layer.name} from dask → numpy (to enable painting)")
                layer.data = layer.data.compute()

        self.viewer.layers.events.inserted.connect(ensure_numpy_on_add)

    def refresh_image_layers(self):
        # Save current selection
        current = self.image_layer_combo.currentText()
        self.image_layer_combo.clear()

        # Only show image layers
        image_layers = [layer.name for layer in self.viewer.layers if isinstance(layer, napari.layers.Image)]
        self.image_layer_combo.addItems(image_layers)

        # Try to restore previous selection if possible
        if current in image_layers:
            idx = image_layers.index(current)
            self.image_layer_combo.setCurrentIndex(idx)

    def change_model(self):
        selected_model = self.model_combo.currentText()
        plugin_dir = os.path.dirname(__file__)
        model_basedir = os.path.join(plugin_dir, "resources", "models")
        self.model = StarDist2D(
            None,
            name=selected_model,
            basedir=model_basedir
        )
        # Update thresholds spin boxes with new model thresholds
        self.wells_prob_spin.setValue(self.model.thresholds1['prob'])
        self.wells_overlap_spin.setValue(self.model.thresholds1['nms'])
        self.plaques_prob_spin.setValue(self.model.thresholds2['prob'])
        self.plaques_overlap_spin.setValue(self.model.thresholds2['nms'])
        # Update resize spin boxes with model's input size
        input_size = getattr(self.model.config, "input_size", [800, 600])
        self.target_width_spin.setValue(input_size[0])
        self.target_height_spin.setValue(input_size[1])

    def reset_all_thresholds(self):
        """Reset all thresholds to the model's original values."""
        self.wells_prob_spin.setValue(self.model.thresholds1['prob'])
        self.wells_overlap_spin.setValue(self.model.thresholds1['nms'])
        self.plaques_prob_spin.setValue(self.model.thresholds2['prob'])
        self.plaques_overlap_spin.setValue(self.model.thresholds2['nms'])

    def reset_wells_thresholds(self):
        """Reset only Wells thresholds to the model's original values."""
        self.wells_prob_spin.setValue(self.model.thresholds1['prob'])
        self.wells_overlap_spin.setValue(self.model.thresholds1['nms'])

    def reset_plaques_thresholds(self):
        """Reset only Plaques thresholds to the model's original values."""
        self.plaques_prob_spin.setValue(self.model.thresholds2['prob'])
        self.plaques_overlap_spin.setValue(self.model.thresholds2['nms'])

    def _get_well_info(self, wells_data, plaque_data, well_diameter_mm=None):
        """
        Helper to extract well binning, plaque counts, average areas, and well diameters from label arrays.
        Returns:
            counts: list of plaque counts per well (6 bins)
            avg_areas: list of average plaque areas per well (6 bins)
            avg_well_diameter_px: average well diameter in pixels
            px_per_cm: pixel-to-cm scale (using 3.5 cm well diameter)
        """
        if well_diameter_mm is None:
            well_diameter_mm = self.WELL_DIAMETER_MM
        # Convert Dask arrays to numpy if needed
        if isinstance(wells_data, da.Array):
            wells_data = wells_data.compute()
        if isinstance(plaque_data, da.Array):
            plaque_data = plaque_data.compute()
        # Select current frame if stack
        if wells_data.ndim >= 3 and plaque_data.ndim >= 3 and wells_data.shape == plaque_data.shape:
            z_idx = 0
            if hasattr(self.viewer, "dims") and hasattr(self.viewer.dims, "current_step"):
                step = self.viewer.dims.current_step
                if isinstance(step, (tuple, list)) and len(step) > 0:
                    z_idx = step[0]
            z_idx = min(max(z_idx, 0), wells_data.shape[0] - 1)
            wells_2d = wells_data[z_idx]
            plaque_2d = plaque_data[z_idx]
        else:
            wells_2d = wells_data
            plaque_2d = plaque_data
        
        # Get unique well ids and their centers
        well_centers = get_well_centers(wells_2d)
        if not well_centers:
            return [0]*6, [0.0]*6, 0.0, 0.0

        # Sort wells into grid
        ordered_wells = sort_wells_grid(well_centers)

        # Calculate stats
        plaque_counts, avg_areas = calculate_well_stats(wells_2d, plaque_2d, ordered_wells)

        # Compute average well diameter in pixels
        avg_well_diameter_px = calculate_well_diameters(wells_2d, well_centers)

        px_per_cm = avg_well_diameter_px / well_diameter_mm if well_diameter_mm > 0 else 0.0
        return plaque_counts, avg_areas, avg_well_diameter_px, px_per_cm

    def count_plaque(self):
        selected_name = self.image_layer_combo.currentText()
        plaque_layer_name = f"{selected_name} Plaque"
        wells_layer_name = f"{selected_name} Wells"
        if plaque_layer_name not in self.viewer.layers or wells_layer_name not in self.viewer.layers:
            # Clear counts if layers are missing
            for label in self.well_count_labels:
                label.setText("")
            return
        plaque_labels = self.viewer.layers[plaque_layer_name].data
        wells_labels = self.viewer.layers[wells_layer_name].data
        counts, _, _, _ = self._get_well_info(wells_labels, plaque_labels)
        # Update labels
        for i, label in enumerate(self.well_count_labels):
            label.setText(str(counts[i]) if counts else "")

    def _connect_dims_slider(self):
        # Connect the dims slider event to update counts in real-time
        if hasattr(self.viewer, "dims") and hasattr(self.viewer.dims, "events"):
            # Disconnect previous if already connected
            try:
                self.viewer.dims.events.current_step.disconnect(self.count_plaque)
            except Exception:
                pass
            self.viewer.dims.events.current_step.connect(self.count_plaque)



    def predict_single(self, img2d):
        target_width = int(self.target_width_spin.value())
        target_height = int(self.target_height_spin.value())
        h0, w0 = img2d.shape[:2]
        scale = min(w0 / target_width, h0 / target_height)
        new_w = int(w0 / scale)
        new_h = int(h0 / scale)
        y_indices = np.linspace(0, h0 - 1, new_h).astype(int)
        x_indices = np.linspace(0, w0 - 1, new_w).astype(int)
        if img2d.ndim == 3:
            resized = img2d[y_indices[:, None], x_indices[None, :], :]
        else:
            resized = img2d[y_indices[:, None], x_indices[None, :]]
        img = img_as_float32(resized)
        axis_norm = (0, 1)
        img = normalize(img, 1, 99.8, axis=axis_norm)
        labels1, labels2 = self.model.predict_instances(
            img,
            prob_thresh1=self.wells_prob_spin.value(),
            prob_thresh2=self.plaques_prob_spin.value(),
            nms_thresh1=self.wells_overlap_spin.value(),
            nms_thresh2=self.plaques_overlap_spin.value()
        )
        labels1[1]['coord'] = [((coord[0] * scale).astype(int), (coord[1] * scale).astype(int)) for coord in labels1[1]['coord']]
        labels2[1]['coord'] = [((coord[0] * scale).astype(int), (coord[1] * scale).astype(int)) for coord in labels2[1]['coord']]
        labels1[1]['points'] = [((pt[0] * scale).astype(int), (pt[1] * scale).astype(int)) for pt in labels1[1]['points']]
        labels2[1]['points'] = [((pt[0] * scale).astype(int), (pt[1] * scale).astype(int)) for pt in labels2[1]['points']]
        return labels1[1], labels2[1]

    def run_prediction(self):
        selected_name = self.image_layer_combo.currentText()
        if not selected_name:
            print("No image layer selected.")
            return
        # Check if selected layer still exists in the viewer
        if selected_name not in self.viewer.layers:
            print(f"Selected layer '{selected_name}' not found in viewer. Refreshing layer list.")
            self.refresh_image_layers()
            return
        image_layer = self.viewer.layers[selected_name]
        # Convert Dask array to numpy if needed
        if isinstance(image_layer, Labels) and isinstance(image_layer.data, da.Array):
            image_layer.data = image_layer.data.compute()
        image = image_layer.data

        # Detect if input is a stack: [z, y, x] or [z, y, x, c]
        is_stack = False
        if image.ndim == 3 and image.shape[-1] > 3: 
            is_stack = True
        elif image.ndim == 4 and image.shape[0] > 1:
            is_stack = True

        # Get UI resize targets
        target_width = int(self.target_width_spin.value())
        target_height = int(self.target_height_spin.value())
        
        # Prepare colormaps
        white_cmap, blue_cmap = create_hydra_colormaps()

        wells_layer_name = f"{selected_name} Wells"
        plaque_layer_name = f"{selected_name} Plaque"

        # For stacks: add frames incrementally to label layers
        if is_stack:
            n_frames = image.shape[0]
            h, w = image.shape[1], image.shape[2]
            # Create empty arrays to accumulate labels
            wells_labels_stack = np.zeros((n_frames, h, w), dtype=np.int32)
            plaque_labels_stack = np.zeros((n_frames, h, w), dtype=np.int32)
            
            # Add/Get layers
            wells_layer = get_or_create_layer(
                self.viewer, wells_layer_name, wells_labels_stack[:1], white_cmap, "multiplicative"
            )
            plaque_layer = get_or_create_layer(
                self.viewer, plaque_layer_name, plaque_labels_stack[:1], blue_cmap, "additive"
            )

            # Set colormap (in case layer existed)
            wells_layer.colormap = white_cmap
            plaque_layer.colormap = blue_cmap
            plaque_layer.contour = 2
            # For each frame, process and update layer data incrementally
            for z in range(n_frames):
                if image.ndim == 4:
                    this_img = image[z]
                else:
                    this_img = image[z]
                if isinstance(this_img, da.Array):
                    this_img = this_img.compute()
                well_labels, plaque_labels = self.predict_single(this_img)
                if z == 0:
                    self.well_labels = well_labels
                    self.plaque_labels = plaque_labels
                wells_labels_arr, _ = rasterize_labels(well_labels, (h, w))
                plaque_labels_arr, _ = rasterize_labels(plaque_labels, (h, w))
                wells_labels_stack[z] = wells_labels_arr
                plaque_labels_stack[z] = plaque_labels_arr
                # Update layers with current stack up to this frame
                wells_layer.data = wells_labels_stack[:z+1].copy()
                plaque_layer.data = plaque_labels_stack[:z+1].copy()
                # Update the dimension slider to the current frame (z axis)
                if 'z' in self.viewer.dims.axis_labels:
                    z_axis = list(self.viewer.dims.axis_labels).index('z')
                    self.viewer.dims.set_current_step(z_axis, z)
                # Force GUI update
                QApplication.processEvents()
            # After all frames, update layers with full stack and set slider to last
            wells_layer.data = wells_labels_stack
            plaque_layer.data = plaque_labels_stack
            if 'z' in self.viewer.dims.axis_labels:
                    z_axis = list(self.viewer.dims.axis_labels).index('z')
                    self.viewer.dims.set_current_step(z_axis, n_frames - 1)
        else:
            h, w = image.shape[:2]
            well_labels, plaque_labels = self.predict_single(image)
            self.well_labels = well_labels
            self.plaque_labels = plaque_labels
            wells_labels_arr, _ = rasterize_labels(well_labels, (h, w))
            plaque_labels_arr, _ = rasterize_labels(plaque_labels, (h, w))
            
            # Add/Get layers
            wells_layer = get_or_create_layer(
                self.viewer, wells_layer_name, wells_labels_arr, white_cmap, "multiplicative"
            )
            plaque_layer = get_or_create_layer(
                self.viewer, plaque_layer_name, plaque_labels_arr, blue_cmap, "additive"
            )

            wells_layer.colormap = white_cmap
            plaque_layer.colormap = blue_cmap
            plaque_layer.contour = 2
        self.count_plaque()
        # Connect dims slider for real-time updating if stack
        self._connect_dims_slider()
            
    def export_prediction(self):
        selected_name = self.image_layer_combo.currentText()
        if not selected_name:
            QMessageBox.warning(self, "Export Prediction", "No image layer selected.")
            return
        # Check if selected layer still exists in the viewer
        if selected_name not in self.viewer.layers:
            QMessageBox.warning(self, "Export Prediction", f"Selected layer '{selected_name}' not found in viewer.")
            self.refresh_image_layers()
            return

        wells_layer_name = f"{selected_name} Wells"
        plaque_layer_name = f"{selected_name} Plaque"
        if wells_layer_name not in self.viewer.layers or plaque_layer_name not in self.viewer.layers:
            QMessageBox.warning(self, "Export Prediction", "Prediction layers not found.")
            return

        wells_data = self.viewer.layers[wells_layer_name].data
        plaque_data = self.viewer.layers[plaque_layer_name].data

        # Ensure data is numpy
        if isinstance(wells_data, da.Array):
            wells_data = wells_data.compute()
        if isinstance(plaque_data, da.Array):
            plaque_data = plaque_data.compute()

        # Handle stacks
        well_diameter_mm = self.WELL_DIAMETER_MM
        if wells_data.ndim == 3:  # stack: [frames, h, w]
            n_frames = wells_data.shape[0]
            counts_per_frame = []
            avg_areas_per_frame = []
            for z in range(n_frames):
                counts, avg_areas, _, px_per_mm = self._get_well_info(wells_data[z], plaque_data[z], well_diameter_mm=well_diameter_mm)
                counts_per_frame.append(counts)
                avg_areas_mm2 = [area / px_per_mm**2 if px_per_mm > 0 else 0.0 for area in avg_areas]
                avg_areas_per_frame.append(avg_areas_mm2)
        else:  # single image
            counts, avg_areas, _, px_per_mm = self._get_well_info(wells_data, plaque_data, well_diameter_mm=well_diameter_mm)
            counts_per_frame = [counts]
            avg_areas_mm2 = [area / px_per_mm**2 if px_per_mm > 0 else 0.0 for area in avg_areas]
            avg_areas_per_frame = [avg_areas_mm2]

        # Prompt for save location
        save_path, _ = QFileDialog.getSaveFileName(
            self, "Save Prediction Summary", f"{selected_name}_summary.txt", "Text Files (*.txt)"
        )
        if not save_path:
            return

        try:
            write_prediction_summary(save_path, well_diameter_mm, px_per_mm, counts_per_frame, avg_areas_per_frame)
            QMessageBox.information(self, "Export Prediction", f"Prediction summary exported to:\n{save_path}")
        except Exception as e:
            QMessageBox.critical(self, "Export Prediction", f"Failed to export summary:\n{e}")

    def tune_model(self):
        """
        Fine-tune the model using the currently selected image and label layers.
        Supports both single images and stacks.
        """
        # LOAD MODEL
        selected_model = self.model_combo.currentText()
        plugin_dir = os.path.dirname(__file__)
        model_basedir = os.path.join(plugin_dir, "resources", "models")
        self.model = StarDist2D(
            None,
            name=selected_model,
            basedir=model_basedir
        )
        config = self.model.config

        # IMAGES & MASKS
        selected_name = self.image_layer_combo.currentText()
        if not selected_name:
            print("No image layer selected.")
            return
        # Check if selected layer still exists in the viewer
        if selected_name not in self.viewer.layers:
            print(f"Selected layer '{selected_name}' not found in viewer. Refreshing layer list.")
            self.refresh_image_layers()
            return
        wells_layer_name = f"{selected_name} Wells"
        plaque_layer_name = f"{selected_name} Plaque"
        if wells_layer_name not in self.viewer.layers or plaque_layer_name not in self.viewer.layers:
            print("Wells or Plaque label layers not found.")
            return
        image_layer = self.viewer.layers[selected_name].data
        wells_layer  = self.viewer.layers[wells_layer_name].data
        plaque_layer = self.viewer.layers[plaque_layer_name].data

        # Convert Dask arrays to numpy
        image = image_layer.compute() if isinstance(image_layer, da.Array) else image_layer
        wells = wells_layer.compute() if isinstance(wells_layer, da.Array) else wells_layer
        plaque = plaque_layer.compute() if isinstance(plaque_layer, da.Array) else plaque_layer

        # Determine if input is a stack
        is_stack = is_image_stack(image)
        
        # If not a stack, wrap single image into a single-frame stack
        if not is_stack:
            # Warn user about single image fine-tuning
            QMessageBox.warning(self, "Single Image Fine-tuning", 
                "Fine-tuning on a single image may result in an overfitted model with poor generalization.")
            # Ensure labels are 2D before wrapping
            img_h, img_w = image.shape[:2]
            img_pixels = img_h * img_w

            
            # Now wrap into single-frame stack
            image = np.expand_dims(image, 0)
            wells = np.expand_dims(wells, 0)
            plaque = np.expand_dims(plaque, 0)
            is_stack = True

        # Get target size from UI
        target_width = int(self.target_width_spin.value())
        target_height = int(self.target_height_spin.value())

        # Prepare batches
        X_train, Y_train = prepare_training_batches(
            image, wells, plaque, is_stack, 
            target_width, target_height, config
        )

        # Get training parameters from UI
        n_frames = image.shape[0]
        batch_size = min(int(self.batch_spin.value()), n_frames) if is_stack else 1
        epochs = int(self.epochs_spin.value())
        learning_rate = float(self.lr_spin.value())

        # Compile and train with specified optimizer and parameters
        optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
        self.model.keras_model.compile(optimizer=optimizer, loss="mse")
        show_info("Tuning the model...")
        self.model.keras_model.fit(X_train, Y_train, epochs=epochs, batch_size=batch_size)
        self.model.keras_model.save_weights(os.path.join(model_basedir, selected_model, "weights_best.h5"))
        self.model.keras_model.load_weights(os.path.join(model_basedir, selected_model, "weights_best.h5"))
        show_info("Model tuned!")
        self.run_prediction()
        show_info("Prediction updated!")

def napari_experimental_provide_dock_widget():
    """This function makes the widget discoverable by napari as a plugin dock widget"""
    return HydraStarDistPlugin


import os
from unittest.mock import MagicMock, patch
from napari_hydra.utils.ensure_default_model import ensure_default_model

@patch('napari_hydra.utils.ensure_default_model.urllib.request.urlretrieve')
@patch('napari_hydra.utils.ensure_default_model.zipfile.ZipFile')
@patch('napari_hydra.utils.ensure_default_model.os.listdir')
@patch('napari_hydra.utils.ensure_default_model.os.path.isdir')
def test_ensure_default_model_download(mock_isdir, mock_listdir, mock_zipfile, mock_urlretrieve):
    # Setup
    dest_dir = "/fake/models"
    zip_url = "http://fake.com/model.zip"
    
    # Case 1: Models already exist
    mock_listdir.return_value = ["existing_model"]
    mock_isdir.return_value = True
    
    dirs = ensure_default_model(dest_dir, zip_url)
    assert "existing_model" in dirs
    mock_urlretrieve.assert_not_called()
    
    # Case 2: No models, download required
    # First call to listdir returns empty (check for existing)
    # Second call (at end) returns downloaded model
    mock_listdir.side_effect = [[], ["downloaded_model"]]
    
    # Mock zipfile extraction
    mock_zip_instance = MagicMock()
    mock_zipfile.return_value.__enter__.return_value = mock_zip_instance
    
    dirs = ensure_default_model(dest_dir, zip_url)
    
    mock_urlretrieve.assert_called_once()
    mock_zip_instance.extractall.assert_called_once_with(dest_dir)
    assert "downloaded_model" in dirs

@patch('napari_hydra.utils.ensure_default_model.urllib.request.urlretrieve')
@patch('napari_hydra.utils.ensure_default_model.shutil.unpack_archive')
@patch('napari_hydra.utils.ensure_default_model.zipfile.ZipFile')
@patch('napari_hydra.utils.ensure_default_model.os.listdir')
def test_ensure_default_model_fallback(mock_listdir, mock_zipfile, mock_unpack, mock_urlretrieve):
    # Test fallback to shutil if BadZipFile
    import zipfile
    dest_dir = "/fake/models"
    zip_url = "http://fake.com/model.zip"
    
    mock_listdir.side_effect = [[], ["unpacked_model"]]
    mock_zipfile.side_effect = zipfile.BadZipFile
    
    ensure_default_model(dest_dir, zip_url)
    
    mock_unpack.assert_called_once()



from napari.utils.colormaps import DirectLabelColormap

def create_hydra_colormaps():
    """
    Create the standard white (wells) and blue (plaque) colormaps used by the plugin.

    Returns:
        tuple: A tuple containing two colormaps:
            - white_cmap (napari.utils.colormaps.DirectLabelColormap): White colormap for well labels.
            - blue_cmap (napari.utils.colormaps.DirectLabelColormap): Blue colormap for plaque labels.
    """
    color_dict = {0: (0, 0, 0, 0)}
    color_dict[None] = (1, 1, 1, 1)
    white_cmap = DirectLabelColormap(color_dict=color_dict)

    color_dict_blue = {0: (0, 0, 0, 0)}
    color_dict_blue[None] = (0, 194/255, 1, 1)
    blue_cmap = DirectLabelColormap(color_dict=color_dict_blue)
    
    return white_cmap, blue_cmap

def get_or_create_layer(viewer, layer_name, layer_data, colormap, blending):
    """
    Get an existing labels layer by name or create a new one with the specified parameters.

    Args:
        viewer (napari.Viewer): The napari viewer instance.
        layer_name (str): Name of the layer to retrieve or create.
        layer_data (numpy.ndarray): Data to initialize the layer with if creating new.
        colormap (napari.utils.colormaps.DirectLabelColormap): Colormap to use for the layer.
        blending (str): Blending mode for the layer (e.g., 'additive', 'multiplicative').

    Returns:
        napari.layers.Labels: The retrieved or newly created labels layer.
    """
    if layer_name in viewer.layers:
        layer = viewer.layers[layer_name]
    else:
        layer = viewer.add_labels(
            layer_data, name=layer_name,
            blending=blending, colormap=colormap
        )
    return layer



import sys
import matplotlib
matplotlib.rcParams["image.interpolation"] = 'none'
import matplotlib.pyplot as plt

import numpy as np
from pathlib import Path
from tqdm import tqdm
from tifffile import imread

from csbdeep.utils import Path, normalize
from stardist import relabel_image_stardist, random_label_cmap, fill_label_holes, calculate_extents, \
    _draw_polygons

from stardist.matching import matching_dataset
from skimage.util import img_as_float32

lbl_cmap = random_label_cmap()

def show_reconstruction_acc(Y, branch, log_dir, results_dir):
    # Fitting ground-truth labels with star-convex polygons
    n_rays = [2**i for i in range(2,8)]
    scores = []
    for r in tqdm(n_rays):
        Y_reconstructed = [relabel_image_stardist(lbl, n_rays=r) for lbl in Y]
        mean_iou = matching_dataset(Y, Y_reconstructed, thresh=0, show_progress=False).mean_true_score
        scores.append(mean_iou)

    plt.figure(figsize=(8,5))
    plt.plot(n_rays, scores, 'o-')
    plt.xlabel('Number of rays for star-convex polygon')
    plt.ylabel('Reconstruction score (mean intersection over union)')
    plt.title("Accuracy of ground truth reconstruction (should be > 0.8 for a reasonable number of rays)")
    path = Path(results_dir+'/'+'/'+log_dir+'/plots/'+'n_rays_reconstruction_accuracy'+str(branch)+'.svg')
    plt.savefig(path)
    None;

def show_reconstruction_polygon(lbl, branch, log_dir,results_dir):
    n_rays = [2**i for i in range(2,8)]
    fig, ax = plt.subplots(2,3, figsize=(16,11))
    for a,r in zip(ax.flat,n_rays):
        a.imshow(relabel_image_stardist(lbl, n_rays=r), cmap=lbl_cmap)
        a.set_title('Reconstructed (%d rays)' % r)
        a.axis('off')
    plt.tight_layout();
    path = Path(results_dir+'/'+'/'+log_dir+'/plots/'+'n_rays_reconstruction_polygon'+str(branch)+'.svg')
    plt.savefig(path)

def load_and_preprocess_data(X, Y):
    X = list(map(imread,X))
    Y = list(map(imread,Y))
    n_channel = 1 if X[0].ndim == 2 else X[0].shape[-1]

    axis_norm = (0,1)   # normalize channels independently
    # axis_norm = (0,1,2) # normalize channels jointly
    if n_channel > 1:
        print("Normalizing image channels %s." % ('jointly' if axis_norm is None or 2 in axis_norm else 'independently'))
        sys.stdout.flush()

    X = [normalize(x,0,100, axis=axis_norm) for x in tqdm(X)]
    Y = [fill_label_holes(y) for y in tqdm(Y)]
    return X, Y

def resize_to_target(image, target_dim):
    def make_divisible(x, divisor=16):
        return (x // divisor) * divisor
    
    h, w = image.shape[:2]
    scale = min(w / target_dim[0], h / target_dim[1])
    new_w = make_divisible(int(w / scale))
    new_h = make_divisible(int(h / scale))
    
    y_indices = np.linspace(0, h - 1, new_h).astype(int)
    x_indices = np.linspace(0, w - 1, new_w).astype(int)
    if image.ndim == 3:
        resized = image[y_indices[:, None], x_indices[None, :], :]
    else:
        resized = image[y_indices[:, None], x_indices[None, :]]
    return resized

def load_and_preprocess_data_hydra(X, Y1, Y2, target_dim=None):
    X = list(map(imread,X))
    Y1 = list(map(imread,Y1))
    Y2 = list(map(imread,Y2))

    if target_dim is not None:
        X = [resize_to_target(x, target_dim) for x in X]
        Y1 = [resize_to_target(y1, target_dim) for y1 in Y1]
        Y2 = [resize_to_target(y2, target_dim) for y2 in Y2]

    n_channel = 1 if X[0].ndim == 2 else X[0].shape[-1]

    axis_norm = (0,1)   # normalize channels independently
    # axis_norm = (0,1,2) # normalize channels jointly
    if n_channel > 1:
        print("Normalizing image channels %s." % ('jointly' if axis_norm is None or 2 in axis_norm else 'independently'))
        sys.stdout.flush()

    X = [normalize(x,0,100, axis=axis_norm) for x in tqdm(X)]
    Y1 = [fill_label_holes(y1) for y1 in tqdm(Y1)]
    Y2 = [fill_label_holes(y2) for y2 in tqdm(Y2)]
    return X, Y1, Y2

def plot_img_label(img, lbl, img_title="image", lbl_title=["label"], **kwargs):
    fig, (ai,aj) = plt.subplots(1,2, figsize=(12,5), gridspec_kw=dict(width_ratios=(1.25,1)))
    im = ai.imshow(img, cmap='gray', clim=(0,1))
    ai.set_title(img_title)    
    fig.colorbar(im, ax=ai)
    aj.imshow(lbl, cmap=lbl_cmap)
    aj.set_title(lbl_title[0])
    plt.tight_layout()

def plot_img_label_hydra(img, lbl1, lbl2, img_title="image", lbl_title=["label1", "label2"], **kwargs):
    fig, (ai,aj,ak) = plt.subplots(1,3, figsize=(12,5), gridspec_kw=dict(width_ratios=(1.25,1,1)))
    im = ai.imshow(img, cmap='gray', clim=(0,1))
    ai.set_title(img_title)    
    fig.colorbar(im, ax=ai)
    aj.imshow(lbl1, cmap=lbl_cmap)
    aj.set_title(lbl_title[0])
    ak.imshow(lbl2, cmap=lbl_cmap)
    ak.set_title(lbl_title[1])
    plt.tight_layout()

def check_fov(Y, model):
    median_size = calculate_extents(list(Y), np.median)
    fov = np.array(model._axes_tile_overlap('YX'))
    print(f"median object size:      {median_size}")
    print(f"network field of view :  {fov}")
    if median_size.any() > fov.any():
        print("WARNING: median object size larger than field of view of the neural network.")

def random_fliprot(img, mask): 
    assert img.ndim >= mask.ndim
    axes = tuple(range(mask.ndim))
    perm = tuple(np.random.permutation(axes))
    img = img.transpose(perm + tuple(range(mask.ndim, img.ndim))) 
    mask = mask.transpose(perm) 
    flipped = False
    for ax in axes: 
        if np.random.rand() > 0.5:
            img = np.flip(img, axis=ax)
            mask = np.flip(mask, axis=ax)
            flipped = True
    return img, mask, flipped

def random_fliprot_hydra(img, mask1, mask2): 
    assert img.ndim >= mask1.ndim == mask2.ndim
    axes = tuple(range(mask1.ndim))
    perm = tuple(np.random.permutation(axes))
    img = img.transpose(perm + tuple(range(mask1.ndim, img.ndim))) 
    mask1 = mask1.transpose(perm) 
    mask2 = mask2.transpose(perm)
    flipped = False
    for ax in axes: 
        if np.random.rand() > 0.5:
            img = np.flip(img, axis=ax)
            mask1 = np.flip(mask1, axis=ax)
            mask2 = np.flip(mask2, axis=ax)
            flipped = True
    return img, mask1, mask2, flipped

def random_intensity_change(img):
    img = img*np.random.uniform(0.6,2) + np.random.uniform(-0.2,0.2)
    return img

def augmenter(x, y):
    """Augmentation of a single input/label image pair.
    x is an input image
    y is the corresponding ground-truth label image
    """
    x, y, flipped = random_fliprot(x, y)
    x = random_intensity_change(x)
    # add some gaussian noise
    sig = 0.02*np.random.uniform(0,1)
    x = x + sig*np.random.normal(0,1,x.shape)
    return x, y

def augmenter_hydra(x, y1, y2):
    """Augmentation of a single input/label image pair.
    x is an input image
    y is the corresponding ground-truth label image
    """
    x, y1, y2, flipped = random_fliprot_hydra(x, y1, y2)
    x = random_intensity_change(x)
    # add some gaussian noise
    sig = 0.02*np.random.uniform(0,1)
    x = x + sig*np.random.normal(0,1,x.shape)
    return x, y1, y2

def plot_metrics_vs_tau(stats,branch,log_dir, results_dir):
    taus = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    fig, (ax1,ax2) = plt.subplots(1,2, figsize=(15,5))

    for m in ('precision', 'recall', 'accuracy', 'f1', 'mean_true_score', 'mean_matched_score', 'panoptic_quality'):
        ax1.plot(taus, [s._asdict()[m] for s in stats], '.-', lw=2, label=m)
    ax1.set_xlabel(r'IoU threshold $\tau$')
    ax1.set_ylabel('Metric value')
    ax1.grid()
    ax1.legend()

    for m in ('fp', 'tp', 'fn'):
        ax2.plot(taus, [s._asdict()[m] for s in stats], '.-', lw=2, label=m)
    ax2.set_xlabel(r'IoU threshold $\tau$')
    ax2.set_ylabel('Number #')
    ax2.grid()
    ax2.legend();

    path = Path(results_dir+'/'+log_dir+'/plots/'+'stats'+str(branch)+'_vs_tau.svg')
    plt.savefig(path)

def plot_individual_channel_predictions(img,labels):
    plt.figure(figsize=(8,8))
    plt.imshow(img if img.ndim==2 else img[...,0], clim=(0,1), cmap='gray')
    plt.imshow(labels, cmap=lbl_cmap, alpha=0.5)
    plt.axis('off');

def plot_individual_channel_predictions_hydra(img,labels):
    plt.figure(figsize=(8,8))
    plt.imshow(img if img.ndim==2 else img[...,0], clim=(0,1), cmap='gray')
    plt.imshow(labels[0], cmap=lbl_cmap, alpha=0.5)
    plt.axis('off');

def example(model, X, Y, i, log_dir, results_dir, show_dist=True):
    axis_norm = (0,1)
    img = normalize(X[i], 0,100, axis=axis_norm)
    labels, details = model.predict_instances(img,prob_thresh=model.thresholds.prob, nms_thresh=model.thresholds.nms)

    plt.figure(figsize=(13,10))
    img_show = img if img.ndim==2 else img[...,0]
    coord, points, prob = details['coord'], details['points'], details['prob']
    plt.subplot(121); plt.imshow(img_show, cmap='gray'); plt.subplot(121).set_title('prediction'); plt.axis('off')
    a = plt.axis()
    _draw_polygons(coord, points, prob, show_dist=show_dist)
    plt.axis(a)
    plt.subplot(122); plt.imshow(img_show, cmap='gray'); plt.subplot(122).set_title('ground truth'); plt.axis('off')
    plt.imshow(Y[i], cmap=lbl_cmap, alpha=0.5)
    plt.tight_layout()
    path = Path(results_dir+'/'+log_dir+'/plots/'+'prediction_overlayed_'+str(i)+'.svg')
    plt.savefig(path)

def example_hydra(model, X, Y1, Y2, i, log_dir, results_dir, show_dist=True):
    axis_norm = (0,1)
    img = normalize(X[i], 0,100, axis=axis_norm)
    labels1, labels2 = model.predict_instances(img,prob_thresh1=model.thresholds1['prob'], prob_thresh2=model.thresholds2['prob'],
                            nms_thresh1=model.thresholds1['nms'], nms_thresh2=model.thresholds2['nms'])

    details1 = labels1[1]
    labels1 = labels1[0]

    details2 = labels2[1]
    labels2 = labels2[0]
    plt.figure(figsize=(13,10))
    img_show = img if img.ndim==2 else img[...,0]
    coord1, points1, prob1 = details1['coord'], details1['points'], details1['prob']
    coord2, points2, prob2 = details2['coord'], details2['points'], details2['prob']
    plt.subplot(131); plt.imshow(img_show, cmap='gray'); plt.subplot(131).set_title('prediction1'); plt.axis('off')
    a = plt.axis()
    _draw_polygons(coord1, points1, prob1, show_dist=show_dist)
    plt.axis(a)
    plt.subplot(132); plt.imshow(img_show, cmap='gray'); plt.subplot(132).set_title('prediction2'); plt.axis('off')
    b = plt.axis()
    _draw_polygons(coord2, points2, prob2, show_dist=show_dist)
    plt.axis(b)
    plt.subplot(133); plt.imshow(img_show, cmap='gray'); plt.subplot(133).set_title('ground truth'); plt.axis('off')
    plt.imshow(Y1[i], cmap=lbl_cmap, alpha=0.5)
    plt.imshow(Y2[i], cmap=lbl_cmap, alpha=0.5)
    plt.tight_layout()
    path = Path(results_dir+'/'+log_dir+'/plots/'+'prediction_overlayed_'+str(i)+'.svg')
    plt.savefig(path)

def create_assimilated_dict(l1,l2,l3,metric):
    assimilated_dict = dict()
    for d1, d2, d3 in zip(l1,l2,l3):
        assimilated_dict[str(d1['thresh'])] = dict()
        assimilated_dict[str(d1['thresh'])][0] = np.round(np.mean([d1[metric],d2[metric],d3[metric]]),4)
        assimilated_dict[str(d1['thresh'])][1] = np.round(np.std([d1[metric],d2[metric],d3[metric]]),4)

    return assimilated_dict


import numpy as np


def split_into_patches(img, target_height, target_width):
    num_blocks = int(np.ceil(img.shape[0]/target_height)), int(np.ceil(img.shape[1]/target_width))

    y_breakpoints = np.linspace(0, img.shape[0]-target_height, num=num_blocks[0], endpoint=True, dtype=int)
    x_breakpoints = np.linspace(0, img.shape[1]-target_width, num=num_blocks[1], endpoint=True, dtype=int)
    breakpoints = np.dstack(np.meshgrid(y_breakpoints, x_breakpoints)).reshape(-1, 2)

    patches = []
    for y, x in breakpoints:
        patches.append(img[y: y+target_height, x: x+target_width])

    return np.array(patches), breakpoints

def combine_patches(patches, breakpoints):
    # patches: shape = (num_patches, patch_height, patch_width, num_channels)
    patch_shape = patches.shape[1], patches.shape[2]
    result_height, result_width = np.max(breakpoints, axis=0) + patch_shape
    result = np.zeros((result_height, result_width, 1)) - np.inf
    for patch, start_coords in zip(patches, breakpoints):
        y_start, x_start = start_coords
        y_end, x_end = start_coords + patch_shape
        result[y_start: y_end, x_start: x_end] = np.maximum(result[y_start: y_end, x_start: x_end], patch)

    return result.astype(np.float32)



from __future__ import absolute_import, print_function

try:
    from pathlib import Path
    Path().expanduser()
except (ImportError,AttributeError):
    from pathlib2 import Path

try:
    import tempfile
    tempfile.TemporaryDirectory
except (ImportError,AttributeError):
    from backports import tempfile

try:
    FileNotFoundError = FileNotFoundError
except NameError:
    FileNotFoundError = IOError


import numpy as np
from unittest.mock import MagicMock, patch
from napari_hydra.utils.process_frame import process_frame

@patch('napari_hydra.utils.process_frame.star_dist')
@patch('napari_hydra.utils.process_frame.edt_prob')
@patch('napari_hydra.utils.process_frame.normalize')
@patch('napari_hydra.utils.process_frame.img_as_float32')
def test_process_frame(mock_as_float, mock_norm, mock_edt, mock_stardist):
    # Inputs
    img2d = np.zeros((100, 100))
    wells2d = np.zeros((100, 100), dtype=int)
    plaque2d = np.zeros((100, 100), dtype=int)
    target_width = 50
    target_height = 50
    config = MagicMock()
    config.grid = (2, 2)
    config.n_rays = 32
    
    # Mock returns
    # normalize returns same shape
    mock_as_float.return_value = np.zeros((32, 32)) # transformed size
    mock_norm.return_value = np.zeros((32, 32))
    
    # edt_prob returns probability map
    mock_edt.return_value = np.zeros((32, 32)) 
    
    # star_dist returns dist map
    mock_stardist.return_value = np.zeros((16, 16, 32)) # subsampled by grid
    
    X, dist1, prob1, dist2, prob2 = process_frame(
        img2d, wells2d, plaque2d, target_width, target_height, config
    )
    
    # Assert output shapes
    # X should be normalized image with channel dim if 2D
    assert X.shape == (32, 32, 1) or X.shape == (32, 32)
    
    # Since we mocked inputs and internal funcs, mainly check calls
    mock_norm.assert_called_once()
    assert mock_edt.call_count == 2 # one for wells, one for plaques
    assert mock_stardist.call_count == 2



import numpy as np
from unittest.mock import MagicMock, patch
from napari_hydra.utils.fine_tune import prepare_training_batches

@patch('napari_hydra.utils.fine_tune.process_frame')
def test_prepare_training_batches_single_image(mock_process):
    # Setup
    img = np.zeros((100, 100))
    wells = np.zeros((100, 100))
    plaque = np.zeros((100, 100))
    is_stack = False
    
    # Mock process_frame return
    # X, dist1, prob1, dist2, prob2
    mock_process.return_value = (
        np.zeros((64, 64, 1)), # X
        np.zeros((32, 32, 32)), # dist1
        np.zeros((32, 32, 1)), # prob1
        np.zeros((32, 32, 32)), # dist2
        np.zeros((32, 32, 1))  # prob2
    )
    
    X_train, Y_train = prepare_training_batches(
        img, wells, plaque, is_stack, 100, 100, MagicMock()
    )
    
    # Expect expanded dims for single image
    assert X_train.shape == (1, 64, 64, 1)
    assert Y_train['dist1'].shape == (1, 32, 32, 32)
    mock_process.assert_called_once()

@patch('napari_hydra.utils.fine_tune.process_frame')
def test_prepare_training_batches_stack(mock_process):
    # Setup stack of 2 frames
    img = np.zeros((2, 100, 100))
    wells = np.zeros((2, 100, 100))
    plaque = np.zeros((2, 100, 100))
    is_stack = True
    
    mock_process.return_value = (
        np.zeros((64, 64, 1)), # X
        np.zeros((32, 32, 32)), # dist1
        np.zeros((32, 32, 1)), # prob1
        np.zeros((32, 32, 32)), # dist2
        np.zeros((32, 32, 1))  # prob2
    )
    
    X_train, Y_train = prepare_training_batches(
        img, wells, plaque, is_stack, 100, 100, MagicMock()
    )
    
    # Expect stack dimension
    assert X_train.shape == (2, 64, 64, 1)
    assert Y_train['dist1'].shape == (2, 32, 32, 32)
    assert mock_process.call_count == 2



from __future__ import print_function, unicode_literals, absolute_import, division

import os
import numpy as np
import json
import collections
import platform
import random
from six.moves import range, zip, map, reduce, filter
from .six import Path

def is_tf_backend():
    from .tf import keras_import
    K = keras_import('backend')
    return K.backend() == 'tensorflow'


def backend_channels_last():
    from .tf import keras_import
    K = keras_import('backend')
    assert K.image_data_format() in ('channels_first','channels_last')
    return K.image_data_format() == 'channels_last'


def move_channel_for_backend(X,channel):
    if backend_channels_last():
        return np.moveaxis(X, channel, -1)
    else:
        return np.moveaxis(X, channel,  1)


def load_json(fpath):
    with open(fpath,'r') as f:
        return json.load(f)


def save_json(data,fpath,**kwargs):
    with open(fpath,'w') as f:
        f.write(json.dumps(data,**kwargs))


def normalize(x, pmin=3, pmax=99.8, axis=None, clip=False, eps=1e-20, dtype=np.float32):
    """Percentile-based image normalization."""

    mi = np.percentile(x,pmin,axis=axis,keepdims=True)
    ma = np.percentile(x,pmax,axis=axis,keepdims=True)
    return normalize_mi_ma(x, mi, ma, clip=clip, eps=eps, dtype=dtype)


def normalize_mi_ma(x, mi, ma, clip=False, eps=1e-20, dtype=np.float32):
    if dtype is not None:
        x   = x.astype(dtype,copy=False)
        mi  = dtype(mi) if np.isscalar(mi) else mi.astype(dtype,copy=False)
        ma  = dtype(ma) if np.isscalar(ma) else ma.astype(dtype,copy=False)
        eps = dtype(eps)

    try:
        import numexpr
        x = numexpr.evaluate("(x - mi) / ( ma - mi + eps )")
    except ImportError:
        x =                   (x - mi) / ( ma - mi + eps )

    if clip:
        x = np.clip(x,0,1)

    return x


def normalize_minmse(x, target):
    """Affine rescaling of x, such that the mean squared error to target is minimal."""
    cov = np.cov(x.flatten(),target.flatten())
    alpha = cov[0,1] / (cov[0,0]+1e-10)
    beta = target.mean() - alpha*x.mean()
    return alpha*x + beta

def _raise(e):
    if isinstance(e, BaseException):
        raise e
    else:
        raise ValueError(e)


# https://docs.python.org/3/library/itertools.html#itertools-recipes
def consume(iterator):
    collections.deque(iterator, maxlen=0)


def compose(*funcs):
    return lambda x: reduce(lambda f,g: g(f), funcs, x)

def download_and_extract_zip_file(url, targetdir='.', verbose=True):
    import csv
    from six.moves.urllib.request import urlretrieve
    from six.moves.urllib.parse import urlparse
    from zipfile import ZipFile

    res = urlparse(url)
    if res.scheme in ('','file'):
        url = Path(res.path).resolve().as_uri()
        # local file, 'urlretrieve' will not make a copy
        # -> don't delete 'downloaded' file
        delete = False
    else:
        delete = True

    # verbosity levels:
    # - 0: no messages
    # - 1: status messages
    # - 2: status messages and list of all files
    if isinstance(verbose,bool):
        verbose *= 2

    log = (print) if verbose else (lambda *a,**k: None)

    targetdir = Path(targetdir)
    if not targetdir.is_dir():
        targetdir.mkdir(parents=True,exist_ok=True)

    provided = []

    def content_is_missing():
        try:
            filepath, http_msg = urlretrieve(url+'.contents')
            with open(filepath,'r') as contents_file:
                contents = list(csv.reader(contents_file,delimiter='\t'))
        except:
            return True
        finally:
            if delete:
                try: os.unlink(filepath)
                except: pass

        for size, relpath in contents:
            size, relpath = int(size.strip()), relpath.strip()
            entry = targetdir / relpath
            if not entry.exists():
                return True
            if entry.is_dir():
                if not relpath.endswith('/'): return True
            elif entry.is_file():
                if relpath.endswith('/') or entry.stat().st_size != size: return True
            else:
                return True
            provided.append(relpath)

        return False


    if content_is_missing():
        try:
            log('Files missing, downloading...',end='')
            filepath, http_msg = urlretrieve(url)
            with ZipFile(filepath,'r') as zip_file:
                log(' extracting...',end='')
                zip_file.extractall(str(targetdir))
                provided = zip_file.namelist()
            log(' done.')
        finally:
            if delete:
                try: os.unlink(filepath)
                except: pass
    else:
        log('Files found, nothing to download.')

    if verbose > 1:
        log('\n'+str(targetdir)+':')
        consume(map(lambda x: log('-',Path(x)), provided))

def axes_check_and_normalize(axes,length=None,disallowed=None,return_allowed=False):
    """
    S(ample), T(ime), C(hannel), Z, Y, X
    """
    allowed = 'STCZYX'
    axes is not None or _raise(ValueError('axis cannot be None.'))
    axes = str(axes).upper()
    consume(a in allowed or _raise(ValueError("invalid axis '%s', must be one of %s."%(a,list(allowed)))) for a in axes)
    disallowed is None or consume(a not in disallowed or _raise(ValueError("disallowed axis '%s'."%a)) for a in axes)
    consume(axes.count(a)==1 or _raise(ValueError("axis '%s' occurs more than once."%a)) for a in axes)
    length is None or len(axes)==length or _raise(ValueError('axes (%s) must be of length %d.' % (axes,length)))
    return (axes,allowed) if return_allowed else axes


def axes_dict(axes):
    """
    from axes string to dict
    """
    axes, allowed = axes_check_and_normalize(axes,return_allowed=True)
    return { a: None if axes.find(a) == -1 else axes.find(a) for a in allowed }


def move_image_axes(x, fr, to, adjust_singletons=False):
    """
    x: ndarray
    fr,to: axes string (see `axes_dict`)
    """
    fr = axes_check_and_normalize(fr, length=x.ndim)
    to = axes_check_and_normalize(to)

    fr_initial = fr
    x_shape_initial = x.shape
    adjust_singletons = bool(adjust_singletons)
    if adjust_singletons:
        # remove axes not present in 'to'
        slices = [slice(None) for _ in x.shape]
        for i,a in enumerate(fr):
            if (a not in to) and (x.shape[i]==1):
                # remove singleton axis
                slices[i] = 0
                fr = fr.replace(a,'')
        x = x[tuple(slices)]
        # add dummy axes present in 'to'
        for i,a in enumerate(to):
            if (a not in fr):
                # add singleton axis
                x = np.expand_dims(x,-1)
                fr += a

    if set(fr) != set(to):
        _adjusted = '(adjusted to %s and %s) ' % (x.shape, fr) if adjust_singletons else ''
        raise ValueError(
            'image with shape %s and axes %s %snot compatible with target axes %s.'
            % (x_shape_initial, fr_initial, _adjusted, to)
        )

    ax_from, ax_to = axes_dict(fr), axes_dict(to)
    if fr == to:
        return x
    return np.moveaxis(x, [ax_from[a] for a in fr], [ax_to[a] for a in fr])

def choice(population, k=1, replace=True):
    ver = platform.sys.version_info
    if replace and (ver.major,ver.minor) in [(2,7),(3,5)]: # python 2.7 or 3.5
        # slow if population is large and not a np.ndarray
        return list(np.random.choice(population, k, replace=replace))
    else:
        try:
            # save state of 'random' and set seed using 'np.random'
            state = random.getstate()
            random.seed(np.random.randint(np.iinfo(int).min, np.iinfo(int).max))
            if replace:
                # sample with replacement
                return random.choices(population, k=k)
            else:
                # sample without replacement
                return random.sample(population, k=k)
        finally:
            # restore state of 'random'
            random.setstate(state)


from abc import ABC
from pathlib import Path
from typing import Iterable, Optional, Union

import numpy as np
from albumentations.core.serialization import Serializable as Transform
from tensorflow.keras.utils import Sequence, load_img, to_categorical


class ImageAbstractDataset(Sequence, ABC):

    def __init__(
        self,
        img_dir: Union[str, Path],
        transform: Transform = None,
        batch_size: int = 50,
        seed: Union[None, int, np.random.Generator] = None,
        **kwargs,
    ) -> None:
        """ Initializes the dataset.

        Args:
            img_dir (Union[str, Path]): Directory containing images to use.
            transform (Transform, optional): Albumentations transforms to apply to images.
            batch_size (int, optional): Size of each batch. Defaults to 50.
            epoch_size (int, optional): Total number of samples in each epoch. Defaults to 1000.
        """
        super().__init__(**kwargs)
        self._rng = np.random.default_rng(seed)

        self._img_list = list(sorted(Path(img_dir).rglob("*.tif")))

        self._transform = transform

        self.img_names = [img.name for img in self._img_list]
        self.batch_size = batch_size

    def _generate_data(self, batch_indices: Iterable[int]) -> np.ndarray:
        """ Given indices of images, generates a batch of samples from them

        Args:
            batch_indices (Iterable[int]): Indices of images to be included in the batch

        Returns:
            tuple[np.ndarray, np.ndarray]: Tuple of arrays representing images and masks
        """
        images = []
        for img_idx in batch_indices:
            img = load_img(self._img_list[img_idx], color_mode="grayscale")
            img = np.expand_dims(img, -1)
            if self._transform is not None:
                transformed = self._transform(image=img)
                img = transformed["image"]
            images.append(img)

        return np.stack(images).astype(np.float32)

class ImageRandomDataset(ImageAbstractDataset):
    """ A dataset containing microscopy images.

        Returns a random subset of samples in each batch.
    """

    def __init__(
        self,
        img_dir: Union[str, Path],
        transform: Optional[Transform] = None,
        batch_size: int = 50,
        epoch_size: int = 1000,
        seed: Union[None, int, np.random.Generator] = None,
        **kwargs,
    ) -> None:
        """ Initializes the dataset.

        Args:
            img_dir (Union[str, Path]): Directory containing images to use.
            transform (Transform, optional): Albumentations transforms to apply to images and masks.
            batch_size (int, optional): Size of each batch. Defaults to 50.
            epoch_size (int, optional): Total number of samples in each epoch. Defaults to 1000.
        """
        super().__init__(img_dir, transform, batch_size, seed, **kwargs)

        self.epoch_size = epoch_size

    def __len__(self) -> int:
        return int(np.ceil(self.epoch_size / self.batch_size))

    def __getitem__(self, index: int) -> np.ndarray:
        """ Generates a random batch of images.

            Index argument is not used, it is only an argument because Sequence requires it.
        """
        batch_indices = self._rng.choice(len(self._img_list), size=self.batch_size, replace=True)
        images = self._generate_data(batch_indices)

        return images


from .ensure_default_model import ensure_default_model
from .rasterize_labels import rasterize_labels
from .is_image_stack import is_image_stack
from .make_divisible import make_divisible
from .process_frame import process_frame
from .well_processing import get_well_centers, sort_wells_grid, calculate_well_stats, calculate_well_diameters
from .layer import create_hydra_colormaps, get_or_create_layer
from .fine_tune import prepare_training_batches
from .export import write_prediction_summary


# Helper to get polygons and rasterize
import numpy as np
from skimage.draw import polygon

def rasterize_labels(label_dict, out_shape):
    """
    Convert polygon coordinates from StarDist prediction into a rasterized label array.

    Args:
        label_dict (dict): Dictionary containing prediction results, specifically 'coord'.
                           'coord' is a list of coordinates (arrays of [x, y]).
        out_shape (tuple): Shape of the output array (height, width).

    Returns:
        tuple: A tuple containing two elements:
            - labels_arr (numpy.ndarray): Rasterized integer label mask.
            - poly_coords (list): List of polygon coordinates (list of [x, y] pairs).
    """
    unmatched_coords = label_dict["coord"]
    poly_coords = []
    for coord in unmatched_coords:
        X = coord[0]
        Y = coord[1]
        single_polygon = [[X[i], Y[i]] for i in range(len(X))]
        poly_coords.append(single_polygon)
    labels_arr = np.zeros(out_shape, dtype=np.int32)
    def scale_poly(poly):
        poly = np.array(poly)
        return poly
    for label, poly in enumerate(poly_coords):
        poly_scaled = scale_poly(poly)
        rr, cc = polygon(poly_scaled[:, 0], poly_scaled[:, 1], shape=labels_arr.shape)
        labels_arr[rr, cc] = label + 1
    return labels_arr, poly_coords


from __future__ import print_function, unicode_literals, absolute_import, division
from six.moves import range, zip, map, reduce, filter
from six import string_types

import numpy as np

from .utils import normalize



def plot_history(history,*keys,**kwargs):
    """Plot (Keras) training history returned by :func:`CARE.train`."""
    import matplotlib.pyplot as plt

    logy = kwargs.pop('logy',False)

    if all(( isinstance(k,string_types) for k in keys )):
        w, keys = 1, [keys]
    else:
        w = len(keys)

    plt.gcf()
    for i, group in enumerate(keys):
        plt.subplot(1,w,i+1)
        for k in ([group] if isinstance(group,string_types) else group):
            plt.plot(history.epoch,history.history[k],'.-',label=k,**kwargs)
            if logy:
                plt.gca().set_yscale('log', nonposy='clip')
        plt.xlabel('epoch')
        plt.legend(loc='best')
    plt.show()


def plot_some(*arr, **kwargs):
    """Quickly plot multiple images at once.

    each arr has to be a list of 2D or 3D images

    Example
    =======

    x = np.ones((200,200))
    plot_some([x],[x])

    x = np.ones((5,200,200))
    plot_some(x,x,x)

    """
    title_list = kwargs.pop('title_list',None)
    pmin = kwargs.pop('pmin',0)
    pmax = kwargs.pop('pmax',100)
    cmap = kwargs.pop('cmap','magma')
    imshow_kwargs = kwargs
    return _plot_some(arr=arr, title_list=title_list, pmin=pmin, pmax=pmax, cmap=cmap, **imshow_kwargs)

def _plot_some(arr, title_list=None, pmin=0, pmax=100, cmap='magma', **imshow_kwargs):
    """
    plots a matrix of images

    arr = [ X_1, X_2, ..., X_n]

    where each X_i is a list of images

    :param arr:
    :param title_list:
    :param pmin:
    :param pmax:
    :param imshow_kwargs:
    :return:

    """
    import matplotlib.pyplot as plt
    imshow_kwargs['cmap'] = cmap

    def make_acceptable(a):
        return np.asarray(a)
    def color_image(a):
        return np.stack(map(to_color,a)) if 1 < a.shape[-1] <= 3 else a
    def max_project(a):
        ndim_allowed = 2 + int(1 <= a.shape[-1] <= 3)
        proj_axis = tuple(range(1, 1 + max(0, a[0].ndim - ndim_allowed)))
        return np.max(a, axis=proj_axis)

    arr = map(make_acceptable,arr)
    arr = map(color_image,arr)
    arr = map(max_project,arr)
    arr = list(arr)

    h = len(arr)
    w = len(arr[0])
    plt.gcf()
    for i in range(h):
        for j in range(w):
            plt.subplot(h, w, i * w + j + 1)
            try:
                plt.title(title_list[i][j], fontsize=8)
            except:
                pass
            img = arr[i][j]
            if pmin!=0 or pmax!=100:
                img = normalize(img,pmin=pmin,pmax=pmax,clip=True)
            plt.imshow(np.squeeze(img),**imshow_kwargs)
            plt.axis("off")


def to_color(arr, pmin=1, pmax=99.8, gamma=1., colors=((0, 1, 0), (1, 0, 1), (0, 1, 1))):
    """Converts a 2D or 3D stack to a colored image (maximal 3 channels).

    Parameters
    ----------
    arr : numpy.ndarray
        2D or 3D input data
    pmin : float
        lower percentile, pass -1 if no lower normalization is required
    pmax : float
        upper percentile, pass -1 if no upper normalization is required
    gamma : float
        gamma correction
    colors : list
        list of colors (r,g,b) for each channel of the input

    Returns
    -------
    numpy.ndarray
        colored image
    """
    if not arr.ndim in (2,3):
        raise ValueError("only 2d or 3d arrays supported")

    if arr.ndim ==2:
        arr = arr[np.newaxis]

    ind_min = np.argmin(arr.shape)
    arr = np.moveaxis(arr, ind_min, 0).astype(np.float32)

    out = np.zeros(arr.shape[1:] + (3,))

    eps = 1.e-20
    if pmin>=0:
        mi = np.percentile(arr, pmin, axis=(1, 2), keepdims=True)
    else:
        mi = 0

    if pmax>=0:
        ma = np.percentile(arr, pmax, axis=(1, 2), keepdims=True)
    else:
        ma = 1.+eps

    arr_norm = (1. * arr - mi) / (ma - mi + eps)


    for i_stack, col_stack in enumerate(colors):
        if i_stack >= len(arr):
            break
        for j, c in enumerate(col_stack):
            out[..., j] += c * arr_norm[i_stack]

    return np.clip(out, 0, 1)


def is_image_stack(image):
    """
    Determine if an image array is a stack of frames or a single image.

    This function uses heuristics based on spatial dimensions (height, width) generally being large (>= 50 px)
    and channel/frame dimensions being small (<= 50 px) to distinguish single images from stacks.

    Args:
        image (numpy.ndarray): Input image array of varying dimensionality (2D, 3D, or 4D).

    Returns:
        bool: True if the image is detected as a stack of frames, False if it is a single image.
    """
    if image.ndim == 2:
        # (height, width) - single 2D image
        return False
    elif image.ndim == 3:
        # Could be (h, w, c), (n_frames, h, w), or other
        dims = image.shape
        
        # Check if dims[0] and dims[1] are consecutive spatial dims (both >= 50)
        if dims[0] >= 50 and dims[1] >= 50:
            # dims[2] is likely channels (1-50) or frames (1-N)
            if dims[2] <= 50:
                # Likely (h, w, c) - single image with channels
                return False
            else:
                # dims[2] > 50, likely (h, w, frames) - stack without channels
                return True
        # Check if dims[1] and dims[2] are consecutive spatial dims (both >= 50)
        elif dims[1] >= 50 and dims[2] >= 50:
            # dims[0] is likely frames or channels
            if dims[0] <= 50:
                # Likely (frames/channels, h, w)
                if dims[0] == 1:
                    return False  # Single frame
                else:
                    return True   # Multiple frames
            else:
                # All three dims >= 50, ambiguous - assume (h, w, c)
                return False
        # Check if dims[0] and dims[2] are >= 50 (less likely but possible)
        elif dims[0] >= 50 and dims[2] >= 50:
            # dims[1] is small, likely channels
            if dims[1] <= 50:
                # Likely (h, c, w) or permutation - assume single image
                return False
            else:
                return False
        else:
            # No clear spatial dims >= 50, assume single image
            return False
    elif image.ndim == 4:
        # Could be (n_frames, h, w, c) or (h, w, c, frames) or other
        dims = image.shape
        # Look for two consecutive dims >= 50 (spatial)
        for i in range(len(dims) - 1):
            if dims[i] >= 50 and dims[i+1] >= 50:
                # Found spatial dims at positions i and i+1
                # Check remaining dims: one should be frames, one should be channels
                other_dims = [dims[j] for j in range(4) if j != i and j != i+1]
                # If both other dims are small (<=50), likely frames and channels
                if all(d <= 50 for d in other_dims):
                    # At least one should be > 1 to indicate multiple frames
                    if max(other_dims) > 1:
                        return True
                    else:
                        return False
                return False
        # No consecutive spatial dims found
        return False
    else:
        # image.ndim > 4: assume stack (high-dimensional data)
        return True


def make_divisible(x, divisor=16):
    """
    Round down a number to the nearest multiple of a divisor.

    Args:
        x (int): The number to be rounded.
        divisor (int, optional): The divisor. Defaults to 16.

    Returns:
        int: The largest multiple of divisor that is less than or equal to x.
    """
    return (x // divisor) * divisor



import numpy as np
from napari_hydra.utils.is_image_stack import is_image_stack

def test_is_image_stack_2d():
    # 2D image (h, w) -> False
    img = np.zeros((100, 100))
    assert not is_image_stack(img)

def test_is_image_stack_3d_stack():
    # (frames, h, w) where frames > 50 is unlikely for channels
    # But heuristic says: consecutive spatial dims >= 50
    # (100, 100, 10) -> (h, w, c) -> False (if last dim <= 50)
    
    # Let's test the specific heuristics:
    # 1. (h, w, c) with small c: dims[0]>=50, dims[1]>=50, dims[2]<=50 -> False
    img = np.zeros((100, 100, 3))
    assert not is_image_stack(img)

    # 2. (frames, h, w) with frames > 50: dims[1]>=50, dims[2]>=50, dims[0]>50
    # ambiguous case (60, 60, 60). Code heuristic prioritizes dims[0], dims[1] being large spatial.
    # If dims[2] is also large, it assumes stack (frames) rather than channels.
    img = np.zeros((60, 60, 60))
    assert is_image_stack(img)

    # 3. (frames, h, w) with typical stack: (10, 100, 100)
    # dims[1]>=50, dims[2]>=50, dims[0]<=50 and > 1 -> True
    img = np.zeros((10, 100, 100))
    assert is_image_stack(img)

def test_is_image_stack_4d():
    # (t, z, y, x) -> True
    img = np.zeros((5, 1, 100, 100))
    assert is_image_stack(img)



import numpy as np

def get_well_centers(wells_2d):
    """
    Extract unique well IDs and their centroids from a 2D label array.

    Args:
        wells_2d (numpy.ndarray): 2D integer array containing well labels.

    Returns:
        list: A list of tuples, where each tuple contains (well_id, (center_x, center_y)).
              center_x and center_y are floats representing the centroid coordinates.
    """
    well_ids = np.unique(wells_2d)
    well_ids = well_ids[well_ids != 0]
    well_centers = []
    for wid in well_ids:
        ys, xs = np.where(wells_2d == wid)
        if len(xs) == 0 or len(ys) == 0:
            continue
        center_x = np.mean(xs)
        center_y = np.mean(ys)
        well_centers.append((wid, (center_x, center_y)))
    return well_centers

def sort_wells_grid(well_centers):
    """
    Sort well centers into a 2x3 grid and return a map of bin index to well ID.

    Assumes a standard 6-well plate layout (2 rows x 3 columns).

    Args:
        well_centers (list): List of tuples (well_id, (center_x, center_y)) as returned by get_well_centers.

    Returns:
        dict: A dictionary mapping bin index (0-5) to well ID. 
              Indices map to grid positions:
              0: (0,0), 1: (0,1), 2: (0,2)
              3: (1,0), 4: (1,1), 5: (1,2)
    """
    if not well_centers:
        return {}
        
    # If more than 6 wells, select 6 closest to centroid
    if len(well_centers) > 6:
        centroid_x = np.mean([cx for _, (cx, cy) in well_centers])
        centroid_y = np.mean([cy for _, (cx, cy) in well_centers])
        well_centers = sorted(
            well_centers,
            key=lambda item: np.sqrt((item[1][0] - centroid_x)**2 + (item[1][1] - centroid_y)**2)
        )[:6]

    min_x, max_x = min(cx for _, (cx, _) in well_centers), max(cx for _, (cx, _) in well_centers)
    x_dist = max_x - min_x
    min_y, max_y = min(cy for _, (_, cy) in well_centers), max(cy for _, (_, cy) in well_centers)
    y_dist = max_y - min_y
    ordered_wells = {}
    for wid, (cx, cy) in well_centers:
        x_d = (cx - min_x) / x_dist if x_dist > 0 else 0
        if x_d < 0.25:
            x_i = 0
        elif x_d < 0.75:
            x_i = 1
        else:
            x_i = 2
        y_d = (cy - min_y) / y_dist if y_dist > 0 else 0
        if y_d < 0.5:
            y_i = 0
        else:
            y_i = 1
        bin_idx = x_i + 3 * y_i
        ordered_wells[bin_idx] = wid
    return ordered_wells

def calculate_well_stats(wells_2d, plaque_2d, ordered_wells):
    """
    Calculate plaque counts and average plaque areas for each well in the grid.

    Args:
        wells_2d (numpy.ndarray): 2D integer array containing well labels.
        plaque_2d (numpy.ndarray): 2D integer array containing plaque labels.
        ordered_wells (dict): Dictionary mapping bin index (0-5) to well ID.

    Returns:
        tuple: A tuple containing two lists of length 6:
            - plaque_counts (list): Count of unique plaques within each well bin.
            - avg_areas (list): Average area (in pixels) of plaques within each well bin.
    """
    plaque_counts = [0] * 6
    avg_areas = [0.0] * 6
    for bin_idx in range(6):
        wid = ordered_wells.get(bin_idx, None)
        if wid is None:
            continue
        mask_well = (wells_2d == wid)
        plaque_labels_in_well = plaque_2d[mask_well]
        unique_plaques = np.unique(plaque_labels_in_well)
        unique_plaques = unique_plaques[unique_plaques != 0]
        plaque_counts[bin_idx] = len(unique_plaques)
        # For each plaque, compute area
        areas = []
        for pid in unique_plaques:
            area = np.sum((plaque_2d == pid) & mask_well)
            areas.append(area)
        avg_areas[bin_idx] = np.mean(areas) if areas else 0.0
    return plaque_counts, avg_areas

def calculate_well_diameters(wells_2d, well_centers):
    """
    Calculate the average well diameter in pixels based on the furthest points from the centroid.

    Args:
        wells_2d (numpy.ndarray): 2D integer array containing well labels.
        well_centers (list): List of tuples (well_id, (center_x, center_y)).

    Returns:
        float: The average diameter of the wells in pixels. Returns 0.0 if no wells provided.
    """
    well_diameters = []
    for wid, (cx, cy) in well_centers:
        yx = np.column_stack(np.where(wells_2d == wid))
        if yx.shape[0] == 0:
            continue
        dists = np.sqrt((yx[:, 1] - cx)**2 + (yx[:, 0] - cy)**2)
        diameter = 2 * np.max(dists)
        well_diameters.append(diameter)
    
    avg_well_diameter_px = np.mean(well_diameters) if well_diameters else 0.0
    return avg_well_diameter_px



import numpy as np
from napari_hydra.utils.well_processing import get_well_centers, sort_wells_grid, calculate_well_stats, calculate_well_diameters

def test_get_well_centers():
    # Create a simple 10x10 array with one "well" labeled 1 at center (5,5)
    wells_2d = np.zeros((10, 10), dtype=int)
    wells_2d[4:6, 4:6] = 1 # 2x2 square at center
    
    centers = get_well_centers(wells_2d)
    assert len(centers) == 1
    wid, (cx, cy) = centers[0]
    assert wid == 1
    assert 4 <= cx <= 5
    assert 4 <= cy <= 5

def test_sort_wells_grid():
    # 6 wells in 2x3 grid
    # Row 0: (10,10), (30,10), (50,10)
    # Row 1: (10,30), (30,30), (50,30)
    wells = [
        (1, (10, 10)), (2, (30, 10)), (3, (50, 10)),
        (4, (10, 30)), (5, (30, 30)), (6, (50, 30))
    ]
    # Expected:
    # 0->1, 1->2, 2->3
    # 3->4, 4->5, 5->6
    
    ordered = sort_wells_grid(wells)
    assert ordered[0] == 1
    assert ordered[2] == 3
    assert ordered[5] == 6

def test_calculate_well_stats():
    # 2 wells, one with 1 plaque, one with 0
    wells_2d = np.zeros((20, 20), dtype=int)
    wells_2d[0:10, 0:10] = 1 # well 1 top-left
    
    plaque_2d = np.zeros((20, 20), dtype=int)
    plaque_2d[2:4, 2:4] = 1 # plaque 1 inside well 1
    
    ordered_wells = {0: 1} # Bin 0 is well 1
    
    counts, areas = calculate_well_stats(wells_2d, plaque_2d, ordered_wells)
    
    assert counts[0] == 1
    # Plaque is 2x2 = 4 pixels
    assert areas[0] == 4.0
    
    # Other bins empty
    assert counts[1] == 0

def test_calculate_well_diameters():
    # One well, circular approximation
    wells_2d = np.zeros((10, 10), dtype=int)
    # Center (5,5)
    # Point at (5,8) -> dist 3
    # Point at (5,2) -> dist 3
    # Diameter ~ 6
    wells_2d[5, 8] = 1
    wells_2d[5, 2] = 1
    
    centers = [(1, (5, 5))]
    
    dia = calculate_well_diameters(wells_2d, centers)
    # Max dist is 3, diameter = 2*max_dist = 6
    assert abs(dia - 6.0) < 0.001



from unittest.mock import MagicMock
from napari_hydra.utils.layer import create_hydra_colormaps, get_or_create_layer

def test_create_hydra_colormaps():
    white, blue = create_hydra_colormaps()
    # Check if they are Colormap objects (or similar)
    # Since we can't easily assert on napari types in minimal env without napari installed
    # we just check they returned something not None
    assert white is not None
    assert blue is not None
    # If napari is installed, we could check type(white).__name__ == 'DirectLabelColormap'

def test_get_or_create_layer_existing():
    viewer = MagicMock()
    # Mock dictionary behavior for viewer.layers
    viewer.layers = {"my_layer": "existing_layer_obj"}
    
    layer = get_or_create_layer(viewer, "my_layer", None, None, None)
    assert layer == "existing_layer_obj"
    viewer.add_labels.assert_not_called()

def test_get_or_create_layer_new():
    viewer = MagicMock()
    # Mock empty layers
    viewer.layers = {}
    
    mock_layer = MagicMock()
    viewer.add_labels.return_value = mock_layer
    
    data = [1, 2, 3]
    layer = get_or_create_layer(viewer, "new_layer", data, "cmap", "additive")
    
    assert layer == mock_layer
    viewer.add_labels.assert_called_once_with(
        data, name="new_layer", blending="additive", colormap="cmap"
    )



import numpy as np
from .process_frame import process_frame

def prepare_training_batches(image, wells, plaque, is_stack, target_width, target_height, config):
    """
    Prepare batches for fine-tuning the model from image and label data.

    Processes single images or stacks, resizing and normalizing them, and generating
    targets (distances and probabilities) for the StarDist model.

    Args:
        image (numpy.ndarray): Training image data (single frame or stack).
        wells (numpy.ndarray): Ground truth well labels.
        plaque (numpy.ndarray): Ground truth plaque labels.
        is_stack (bool): Flag indicating if the input data represents a stack of frames.
        target_width (int): Target width for resizing images.
        target_height (int): Target height for resizing images.
        config (object): StarDist configuration object containing grid and n_rays.

    Returns:
        tuple: A tuple containing two elements:
            - X_train (numpy.ndarray): Training images array suitable for Keras model input.
            - Y_train (dict): Dictionary of training targets with keys 'dist1', 'prob1', 'dist2', 'prob2'.
    """
    if is_stack:
        n_frames = image.shape[0]
        # If image has more than 3 dims, select first channel if needed
        Xs, dist1s, prob1s, dist2s, prob2s = [], [], [], [], []
        for z in range(n_frames):
            # Handle 4D (t,z,y,x) or (z,y,x,c) etc
            img2d = image[z]
            wells2d = wells[z]
            plaque2d = plaque[z]
            X, dist1, prob1, dist2, prob2 = process_frame(
                img2d, wells2d, plaque2d, 
                target_width, target_height, config
            )
            Xs.append(X)
            dist1s.append(dist1)
            prob1s.append(prob1)
            dist2s.append(dist2)
            prob2s.append(prob2)
        X_train = np.stack(Xs, axis=0)
        Y_train = {
            'dist1': np.stack(dist1s, axis=0),
            'prob1': np.stack(prob1s, axis=0),
            'dist2': np.stack(dist2s, axis=0),
            'prob2': np.stack(prob2s, axis=0),
        }
    else:
        if image.ndim == 3 and image.shape[-1] in (1, 3):
            img2d = image
            wells2d = wells
            plaque2d = plaque
        elif image.ndim == 2:
            img2d = image
            wells2d = wells
            plaque2d = plaque
        else:
            # If image is 3D with shape (1, h, w) or similar, squeeze
            img2d = np.squeeze(image)
            wells2d = np.squeeze(wells)
            plaque2d = np.squeeze(plaque)
        X, dist1, prob1, dist2, prob2 = process_frame(img2d, wells2d, plaque2d)
        X_train = np.expand_dims(X, 0)
        Y_train = {
            'dist1': np.expand_dims(dist1, 0),
            'prob1': np.expand_dims(prob1, 0),
            'dist2': np.expand_dims(dist2, 0),
            'prob2': np.expand_dims(prob2, 0),
        }
    return X_train, Y_train



from __future__ import annotations

from abc import ABC
from collections.abc import Sequence
from pathlib import Path
from typing import Iterable, Optional, Union

import numpy as np
from albumentations.core.serialization import Serializable as Transform
from tensorflow.keras.utils import Sequence as KerasSequence
from tensorflow.keras.utils import load_img, to_categorical


def binarize_mask(mask: np.ndarray, max_value: int = 255) -> np.ndarray:
    mask = mask / max_value
    return np.rint(mask).astype(int)

class ImageMaskAbstractDataset(KerasSequence, ABC):

    def __init__(
        self,
        img_dir: Union[str, Path],
        mask_dir: Union[str, Path],
        num_classes: int = 1,
        transform: Optional[Transform] = None,
        batch_size: int = 50,
        seed: Union[None, int, np.random.Generator] = None,
        num_samples: Optional[int] = None,
        **kwargs,
    ) -> None:
        """ Initializes the dataset.

        Args:
            img_dir (Union[str, Path]): Directory containing images to use.
            mask_dir (Union[str, Path]): Directory containing masks corresponding to the images.
            transform (Transform, optional): Albumentations transforms to apply to images and masks.
            batch_size (int, optional): Size of each batch. Defaults to 50.
            epoch_size (int, optional): Total number of samples in each epoch. Defaults to 1000.
            num_samples (int, optional): Number of samples to subsample from the dataset.

        Raises:
            ValueError: If any image does not match to the mask.
        """
        super().__init__(**kwargs)
        self._rng = np.random.default_rng(seed)

        self._img_list = np.sort(list(Path(img_dir).rglob("*.tif")))
        self._mask_list = np.sort(list(Path(mask_dir).rglob("*.tif")))

        if len(self._img_list) != len(self._mask_list):
            raise RuntimeError("The number of images and masks does not match")

        for img_path, mask_path in zip(self._img_list, self._mask_list):
            if not mask_path.stem.startswith(img_path.stem):
                raise ValueError(
                    f"The image {img_path.name} does not match the mask {mask_path.name}."
                )

        if num_samples is not None:
            indices = self._rng.choice(
                np.arange(len(self._img_list)),
                size=num_samples,
                replace=False,
            )
            self._img_list = self._img_list[indices]
            self._mask_list = self._mask_list[indices]

        self._transform = transform

        self._num_classes = num_classes
        self._batch_size = batch_size

    @property
    def num_classes(self) -> int:
        return self._num_classes

    @property
    def batch_size(self) -> int:
        return self._batch_size

    @property
    def img_names(self) -> Sequence[str]:
        return [img.name for img in self._img_list]


    def _generate_data(self, batch_indices: Iterable[int]) -> tuple[np.ndarray, np.ndarray]:
        """ Given indices of images, generates a batch of samples from them

        Args:
            batch_indices (Iterable[int]): Indices of images to be included in the batch

        Returns:
            tuple[np.ndarray, np.ndarray]: Tuple of arrays representing images and masks
        """
        images, masks = [], []
        for img_idx in batch_indices:
            img = load_img(self._img_list[img_idx], color_mode="grayscale")
            mask = load_img(self._mask_list[img_idx], color_mode="grayscale")
            img = np.expand_dims(img, -1)
            mask = np.array(mask)
            if self._transform is not None:
                transformed = self._transform(image=img, mask=mask)
                img = transformed["image"]
                mask = transformed["mask"]
            images.append(img)
            masks.append(mask)

        images_array = np.stack(images).astype(np.float32)
        masks_array = np.stack(masks).astype(np.float32)
        # rescale images to (-1, 1)
        images_array = images_array*2 - 1
        return images_array, masks_array

    def _expand_masks(self, masks):
        if self.num_classes == 1:
            masks = np.expand_dims(masks, -1)
            masks = binarize_mask(masks)
        else:
            masks = to_categorical(masks, num_classes=self.num_classes)

        return masks


class ImageMaskRandomDataset(ImageMaskAbstractDataset):
    """ A dataset containing microscopy images and corresponding masks.

        Returns a random subset of samples in each batch.
    """

    def __init__(
        self,
        img_dir: Union[str, Path],
        mask_dir: Union[str, Path],
        num_classes: int = 1,
        transform: Transform = None,
        batch_size: int = 50,
        epoch_size: int = 1000,
        max_iters: int = 10,
        threshold_obj_perc: float = 0.01,
        seed: Union[None, int, np.random.Generator] = None,
        expand_masks: bool = True,
        num_samples: Optional[int] = None,
        **kwargs,
    ) -> None:
        """ Initializes the dataset.

        Args:
            img_dir (Union[str, Path]): Directory containing images to use.
            mask_dir (Union[str, Path]): Directory containing masks corresponding to the images.
            transform (Transform, optional): Albumentations transforms to apply to images and masks.
            batch_size (int, optional): Size of each batch. Defaults to 50.
            epoch_size (int, optional): Total number of samples in each epoch. Defaults to 1000.
            max_iters (int, optional): Max number of resampling if batches don"t satisfy condition. Defaults to 10.
            threshold_obj_perc (float, optional): Required ROI density for the batch. Defaults to 0.01.
            num_samples (int, optional): Number of samples to subsample from the dataset.

        Raises:
            ValueError: If any image does not match to the mask.
        """
        super().__init__(img_dir, mask_dir, num_classes, transform, batch_size, seed, num_samples, **kwargs)

        self._max_iters = max_iters
        self._threshold_obj_perc = threshold_obj_perc
        self.expand_masks = expand_masks

        self.epoch_size = epoch_size

    def __len__(self):
        return int(np.ceil(self.epoch_size / self.batch_size))

    def _check_threshhold_condition(self, masks: np.ndarray) -> bool:
        """ Given a batch of masks, checks if they satisfy the condition of ROI density
        """
        num_positive = np.sum(masks.astype(bool))
        num_pixels = np.prod(masks.shape)

        return (num_positive / num_pixels) > self._threshold_obj_perc

    def __getitem__(self, index: int):
        """ Generates a random batch of images.

            Index argument is not used, it is only an argument because Sequence requires it.
        """
        for _ in range(self._max_iters):
            batch_indices = self._rng.choice(len(self._img_list), size=self.batch_size, replace=True)
            images, masks = self._generate_data(batch_indices)
            if self._check_threshhold_condition(masks):
                break

        if self.expand_masks:
            masks = self._expand_masks(masks)

        return images.astype(np.float32), masks.astype(np.float32)


class ImageMaskDataset(ImageMaskAbstractDataset):
    """ A dataset containing microscopy images and corresponding masks.

        Returns each image only once.
    """

    def __init__(
        self,
        img_dir: Union[str, Path],
        mask_dir: Union[str, Path],
        num_classes: int = 1,
        transform: Transform = None,
        batch_size: int = 50,
        shuffle: bool = False,
        seed: Union[None, int, np.random.Generator] = None,
        expand_masks: bool = True,
        num_samples: Optional[int] = None,
        **kwargs,
    ) -> None:
        """ Initializes the dataset.

        Args:
            img_dir (Union[str, Path]): Directory containing images to use.
            mask_dir (Union[str, Path]): Directory containing masks corresponding to the images.
            transform (Transform, optional): Albumentations transforms to apply to images and masks.
            batch_size (int, optional): Size of each batch. Defaults to 50.
            shuffle (bool, optional): Whether to reshuffle the dataset after each epoch. Defaults to False.
        Raises:
            ValueError: If any image does not match to the mask.
        """
        super().__init__(img_dir, mask_dir, num_classes, transform, batch_size, seed, num_samples, **kwargs)

        self._shuffle = shuffle
        self.expand_masks = expand_masks

        self.on_epoch_end()

    def __len__(self):
        return int(np.ceil(len(self._img_list) / self.batch_size))

    def on_epoch_end(self):
        num_images = len(self._img_list)
        if self._shuffle:
            self.indices = self._rng.permutation(num_images)
        else:
            self.indices = np.arange(num_images)

    def __getitem__(self, index: int):
        """ Returns a batch of images.
        """
        batch_indices = self.indices[index*self.batch_size: (index+1)*self.batch_size]
        images, masks = self._generate_data(batch_indices)

        if self.expand_masks:
            masks = self._expand_masks(masks)

        return images.astype(np.float32), masks.astype(np.float32)


class ImageMaskStaticDataset(KerasSequence):
    """ Dataset that always returns the same patches, can be used for plotting
        the outputs of diffrent models.
    """

    def __init__(
        self,
        npz_file_dir,
        ) -> None:
        super().__init__()
        self.npz_file_dir = Path(npz_file_dir)

    def __getitem__(self, index: int) -> tuple[np.ndarray, np.ndarray]:
        data = np.load(self.npz_file_dir / f"batch_{index}.npz")
        return data["images"], data["masks"]

    def __len__(self) -> int:
        return len(list(self.npz_file_dir.iterdir()))

    @staticmethod
    def create_from_directory(
        img_dir: Union[str, Path],
        mask_dir: Union[str, Path],
        npz_file_dir: Union[str, Path],
        num_classes: int = 1,
        transform: Transform = None,
        batch_size: int = 50,
        epoch_size: int = 1000,
        max_iters: int = 10,
        threshold_obj_perc: float = 0.01,
        seed: Union[None, int, np.random.Generator] = None,
        expand_masks: bool = True,
        num_samples: Optional[int] = None,
        **kwargs,
    ) -> ImageMaskStaticDataset:

        npz_file_dir = Path(npz_file_dir)
        random_dataset = ImageMaskRandomDataset(
            img_dir,
            mask_dir,
            num_classes,
            transform,
            batch_size,
            epoch_size,
            max_iters,
            threshold_obj_perc,
            seed,
            expand_masks,
            num_samples,
            **kwargs,
        )

        for batch, (imgs, masks) in enumerate(random_dataset):
            np.savez(npz_file_dir / f"batch_{batch}.npz", images=imgs, masks=masks)

        return ImageMaskStaticDataset(npz_file_dir)



import os
import urllib
import tempfile
import zipfile
import shutil
from napari.utils.notifications import show_info

def ensure_default_model(dest_dir, zip_url):
    """
    Ensure that a default model exists in the destination directory.
    
    If no models are found in `dest_dir`, it attempts to download a default model
    from the specified `zip_url` and extract it.

    Args:
        dest_dir (str): Directory where models should be located.
        zip_url (str): URL to the default model zip file.

    Returns:
        list: A list of subdirectory names (models) found in `dest_dir` after the operation.
    """
    # Check for any directory in models folder
    dirs = [name for name in os.listdir(dest_dir) if os.path.isdir(os.path.join(dest_dir, name))]
    if dirs:
        return dirs
    # Download into a temp file then extract
    try:
        show_info("No models found locally — attempting to download default model (this may take a while)...")
    except Exception:
        pass
    try:
        fd, tmp_path = tempfile.mkstemp(suffix=".zip")
        os.close(fd)
        # download
        urllib.request.urlretrieve(zip_url, tmp_path)
        # extract
        try:
            with zipfile.ZipFile(tmp_path, 'r') as zf:
                zf.extractall(dest_dir)
        except zipfile.BadZipFile:
            # fallback to shutil.unpack_archive which may handle other formats
            shutil.unpack_archive(tmp_path, dest_dir)
        finally:
            try:
                os.remove(tmp_path)
            except Exception:
                pass
    except Exception as e:
        try:
            show_info(f"Failed to download default model: {e}")
        except Exception:
            pass
    # return any directories found after attempted download
    return [name for name in os.listdir(dest_dir) if os.path.isdir(os.path.join(dest_dir, name))]



from napari_hydra.utils.make_divisible import make_divisible

def test_make_divisible():
    assert make_divisible(17, 16) == 16
    assert make_divisible(32, 16) == 32
    assert make_divisible(15, 16) == 0
    assert make_divisible(100, 10) == 100
    assert make_divisible(105, 10) == 100



import os
from napari_hydra.utils.export import write_prediction_summary

def test_write_prediction_summary(tmp_path):
    save_path = tmp_path / "summary.txt"
    
    well_diameter_mm = 35.0
    px_per_mm = 10.0
    
    counts_per_frame = [[10, 20, 30, 40, 50, 60]]
    avg_areas_per_frame = [[5.0, 6.0, 7.0, 8.0, 9.0, 10.0]]
    
    write_prediction_summary(
        str(save_path), 
        well_diameter_mm, 
        px_per_mm, 
        counts_per_frame, 
        avg_areas_per_frame
    )
    
    assert save_path.exists()
    content = save_path.read_text()
    
    assert "PREDICTION SUMMARY" in content
    assert "Assumed Well Diameter: 35.0 mm" in content
    assert "Pixel-to-mm Scale: 10.00 px/mm" in content
    assert "FRAME 1" in content
    assert "| 010 | 020 | 030 |" in content
    assert "| 040 | 050 | 060 |" in content



import numpy as np
from skimage import img_as_float32
from csbdeep.utils import normalize
from stardist.utils import edt_prob
from stardist.geometry import star_dist
from .make_divisible import make_divisible

def process_frame(img2d, wells2d, plaque2d, target_width, target_height, config):
    """
    Process a single frame for training, resizing and generating StarDist targets.

    Args:
        img2d (numpy.ndarray): Input 2D image (or 2D + channels).
        wells2d (numpy.ndarray): 2D well labels.
        plaque2d (numpy.ndarray): 2D plaque labels.
        target_width (int): Target width for resizing.
        target_height (int): Target height for resizing.
        config (object): StarDist configuration object containing grid and n_rays.

    Returns:
        tuple: A tuple containing five elements:
            - X (numpy.ndarray): Normalized and resized image with channel dimension.
            - dist1 (numpy.ndarray): StarDist distances for wells.
            - prob1 (numpy.ndarray): StarDist probabilities for wells.
            - dist2 (numpy.ndarray): StarDist distances for plaques.
            - prob2 (numpy.ndarray): StarDist probabilities for plaques.
    """
    # Resize frame and labels
    h, w = img2d.shape[:2]
    scale = min(w / target_width, h / target_height)
    new_w, new_h = int(w / scale), int(h / scale)
    new_w, new_h = make_divisible(new_w), make_divisible(new_h)
    y_indices = np.linspace(0, h - 1, new_h).astype(int)
    x_indices = np.linspace(0, w - 1, new_w).astype(int)

    # Resize image
    if img2d.ndim == 3:
        resized_img = img2d[y_indices[:, None], x_indices[None, :], :]
    else:
        resized_img = img2d[y_indices[:, None], x_indices[None, :]]
    img = img_as_float32(resized_img)
    axis_norm = (0, 1)
    image_norm = normalize(img, 1, 99.8, axis=axis_norm)

    # Resize labels
    wells_resized = wells2d[y_indices[:, None], x_indices[None, :]]
    plaque_resized = plaque2d[y_indices[:, None], x_indices[None, :]]

    # Ensure input has channel axis
    X = image_norm.astype(np.float32)
    if X.ndim == 2:
        X = np.expand_dims(X, -1)
    Y1 = wells_resized.astype(np.int32)
    Y2 = plaque_resized.astype(np.int32)

    # Prepare targets
    prob1_full = edt_prob(Y1)
    prob2_full = edt_prob(Y2)

    y_indices_prob = np.arange(0, prob1_full.shape[0], config.grid[0])
    x_indices_prob = np.arange(0, prob1_full.shape[1], config.grid[1])
    prob1 = np.expand_dims(prob1_full[y_indices_prob[:, None], x_indices_prob[None, :]], -1)
    prob2 = np.expand_dims(prob2_full[y_indices_prob[:, None], x_indices_prob[None, :]], -1)

    dist1 = star_dist(Y1, config.n_rays, mode="cpp", grid=config.grid)
    dist2 = star_dist(Y2, config.n_rays, mode="cpp", grid=config.grid)

    return X, dist1, prob1, dist2, prob2



from __future__ import print_function, unicode_literals, absolute_import, division
from .utils import _raise, is_tf_backend, save_json, backend_channels_last, normalize
from six.moves import range, zip, map, reduce, filter

import numpy as np
import os
from operator import itemgetter
import warnings
import shutil
import datetime
from importlib import import_module
from packaging import version

from tensorflow import __version__ as _tf_version
_tf_version = version.parse(_tf_version)
IS_TF_1 = _tf_version.major == 1
_KERAS = 'keras' if IS_TF_1 else 'tensorflow.keras'

if IS_TF_1:
    try:
        import keras
    except ModuleNotFoundError:
        raise RuntimeError("""

For 'tensorflow' 1.x (found version {_tf_version}), the stand-alone 'keras' package (2.1.2 <= version < 2.4) is required.
When using the most recent version of 'tensorflow' 1.x, install 'keras' like this: pip install "keras>=2.1.2,<2.4"

If you must use an older version of 'tensorflow' 1.x, please see this file for compatible versions of 'keras':
https://github.com/CSBDeep/CSBDeep/blob/master/.github/workflows/tests_legacy.yml
""".format(_tf_version=_tf_version))

elif _tf_version >= version.parse('2.6'):
    try:
        from keras import __version__ as _keras_version
        _keras_version = version.parse(_keras_version)
        if (_tf_version.major,_tf_version.minor) > (_keras_version.major,_keras_version.minor):
            raise RuntimeError("""

Found version {_keras_version} of 'keras', which appears to be too old for the installed version {_tf_version} of 'tensorflow'.
Please update 'keras': pip install "keras>={_tf_version.major}.{_tf_version.minor}"
""".format(_keras_version=_keras_version, _tf_version=_tf_version))
    except ModuleNotFoundError:
        raise RuntimeError("""

Starting with 'tensorflow' version 2.6.0, a recent version of the stand-alone 'keras' package is required.
Please install/update 'keras': pip install "keras>={_tf_version.major}.{_tf_version.minor}"
""".format(_tf_version=_tf_version))


def keras_import(sub=None, *names):
    if sub is None:
        return import_module(_KERAS)
    else:
        mod = import_module('{_KERAS}.{sub}'.format(_KERAS=_KERAS,sub=sub))
        if len(names) == 0:
            return mod
        elif len(names) == 1:
            return getattr(mod, names[0])
        return tuple(getattr(mod, name) for name in names)

import tensorflow as tf

keras = keras_import()
K = keras_import('backend')
Callback = keras_import('callbacks', 'Callback')
Lambda = keras_import('layers', 'Lambda')


from .six import tempfile



def limit_gpu_memory(fraction, allow_growth=False, total_memory=None):
    """Limit GPU memory allocation for TensorFlow (TF) backend.

    Parameters
    ----------
    fraction : float
        Limit TF to use only a fraction (value between 0 and 1) of the available GPU memory.
        Reduced memory allocation can be disabled if fraction is set to ``None``.
    allow_growth : bool, optional
        If ``False`` (default), TF will allocate all designated (see `fraction`) memory all at once.
        If ``True``, TF will allocate memory as needed up to the limit imposed by `fraction`; this may
        incur a performance penalty due to memory fragmentation.
    total_memory :  int or iterable of int
        Total amount of available GPU memory (in MB).

    Raises
    ------
    ValueError
        If `fraction` is not ``None`` or a float value between 0 and 1.
    NotImplementedError
        If TensorFlow is not used as the backend.
    """

    is_tf_backend() or _raise(NotImplementedError('Not using tensorflow backend.'))
    fraction is None or (np.isscalar(fraction) and 0<=fraction<=1) or _raise(ValueError('fraction must be between 0 and 1.'))

    if IS_TF_1:
        _session = None
        try:
            _session = K.tensorflow_backend._SESSION
        except AttributeError:
            pass

        if _session is None:
            config = tf.ConfigProto()
            if fraction is not None:
                config.gpu_options.per_process_gpu_memory_fraction = fraction
            config.gpu_options.allow_growth = bool(allow_growth)
            session = tf.Session(config=config)
            K.tensorflow_backend.set_session(session)
        else:
            warnings.warn('Too late to limit GPU memory, can only be done once and before any computation.')
    else:
        gpus = tf.config.experimental.list_physical_devices('GPU')
        if gpus:
            if fraction is not None:
                np.isscalar(total_memory) or _raise(ValueError("'total_memory' must be provided when using TensorFlow 2."))
                vdc = tf.config.experimental.VirtualDeviceConfiguration(memory_limit=int(np.ceil(total_memory*fraction)))
            try:
                for gpu in gpus:
                    if fraction is not None:
                        tf.config.experimental.set_virtual_device_configuration(gpu,[vdc])
                    if allow_growth:
                        tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # must be set before GPUs have been initialized
                print(e)



def export_SavedModel(model, outpath, meta={}, format='zip'):
    """Export Keras model in TensorFlow's SavedModel_ format.

    See `Your Model in Fiji`_ to learn how to use the exported model with our CSBDeep Fiji plugins.

    .. _SavedModel: https://www.tensorflow.org/programmers_guide/saved_model#structure_of_a_savedmodel_directory
    .. _`Your Model in Fiji`: https://github.com/CSBDeep/CSBDeep_website/wiki/Your-Model-in-Fiji

    Parameters
    ----------
    model : :class:`keras.models.Model`
        Keras model to be exported.
    outpath : str
        Path of the file/folder that the model will exported to.
    meta : dict, optional
        Metadata to be saved in an additional ``meta.json`` file.
    format : str, optional
        Can be 'dir' to export as a directory or 'zip' (default) to export as a ZIP file.

    Raises
    ------
    ValueError
        Illegal arguments.

    """

    def export_to_dir(dirname):
        if len(model.inputs) > 1 or len(model.outputs) > 1:
            warnings.warn('Found multiple input or output layers.')

        def _export(model):
            if IS_TF_1:
                from tensorflow import saved_model
                from keras.backend import get_session
            else:
                from tensorflow.compat.v1 import saved_model
                from tensorflow.compat.v1.keras.backend import get_session

            if not IS_TF_1:
                warnings.warn(\
"""
***IMPORTANT NOTE***

You are using 'tensorflow' 2.x, hence it is likely that the exported model *will not work*
in associated ImageJ/Fiji plugins (e.g. CSBDeep and StarDist).

If you indeed have problems loading the exported model in Fiji, the current workaround is
to load the trained model in a Python environment with installed 'tensorflow' 1.x and then
export it again. If you need help with this, please read:

https://gist.github.com/uschmidt83/4b747862fe307044c722d6d1009f6183
""")

            builder = saved_model.builder.SavedModelBuilder(dirname)
            # use name 'input'/'output' if there's just a single input/output layer
            inputs  = dict(zip(model.input_names,model.inputs))   if len(model.inputs)  > 1 else dict(input=model.input)
            outputs = dict(zip(model.output_names,model.outputs)) if len(model.outputs) > 1 else dict(output=model.output)
            signature = saved_model.signature_def_utils.predict_signature_def(inputs=inputs, outputs=outputs)
            signature_def_map = { saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: signature }
            builder.add_meta_graph_and_variables(get_session(), [saved_model.tag_constants.SERVING], signature_def_map=signature_def_map)
            builder.save()

        if IS_TF_1:
            _export(model)
        else:
            from tensorflow.keras.models import clone_model
            weights = model.get_weights()
            with tf.Graph().as_default():
                # clone model in new graph and set weights
                _model = clone_model(model)
                _model.set_weights(weights)
                _export(_model)

        if meta is not None and len(meta) > 0:
            save_json(meta, os.path.join(dirname,'meta.json'))


    ## checks
    isinstance(model,keras.models.Model) or _raise(ValueError("'model' must be a Keras model."))
    # supported_formats = tuple(['dir']+[name for name,description in shutil.get_archive_formats()])
    supported_formats = 'dir','zip'
    format in supported_formats or _raise(ValueError("Unsupported format '%s', must be one of %s." % (format,str(supported_formats))))

    # remove '.zip' file name extension if necessary
    if format == 'zip' and outpath.endswith('.zip'):
        outpath = os.path.splitext(outpath)[0]

    if format == 'dir':
        export_to_dir(outpath)
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpsubdir = os.path.join(tmpdir,'model')
            export_to_dir(tmpsubdir)
            shutil.make_archive(outpath, format, tmpsubdir)



def tf_normalize(x, pmin=1, pmax=99.8, axis=None, clip=False):
    assert pmin < pmax
    mi = tf.contrib.distributions.percentile(x,pmin, axis=axis, keep_dims=True)
    ma = tf.contrib.distributions.percentile(x,pmax, axis=axis, keep_dims=True)
    y = (x-mi)/(ma-mi+K.epsilon())
    if clip:
        y = K.clip(y,0,1.0)
    return y


def tf_normalize_layer(layer, pmin=1, pmax=99.8, clip=True):
    def norm(x,axis):
        return tf_normalize(x, pmin=pmin, pmax=pmax, axis=axis, clip=clip)

    shape = K.int_shape(layer)
    n_channels_out = shape[-1]
    n_dim_out = len(shape)

    if n_dim_out > 4:
        layer = Lambda(lambda x: K.max(x, axis=tuple(range(1,1+n_dim_out-4))))(layer)

    assert 0 < n_channels_out

    if n_channels_out == 1:
        out = Lambda(lambda x: norm(x, axis=(1,2)))(layer)
    elif n_channels_out == 2:
        out = Lambda(lambda x: norm(K.concatenate([x,x[...,:1]], axis=-1), axis=(1,2,3)))(layer)
    elif n_channels_out == 3:
        out = Lambda(lambda x: norm(x, axis=(1,2,3)))(layer)
    else:
        out = Lambda(lambda x: norm(K.max(x, axis=-1, keepdims=True), axis=(1,2,3)))(layer)
    return out


class CARETensorBoard(Callback):
    """ TODO """
    def __init__(self, log_dir='./logs',
                 freq=1,
                 compute_histograms=False,
                 n_images=3,
                 prob_out=False,
                 write_graph=False,
                 prefix_with_timestamp=True,
                 write_images=False,
                 image_for_inputs=None,      # write images for only these input indices
                 image_for_outputs=None,     # write images for only these output indices
                 input_slices=None,          # list (of list) of slices to apply to `image_for_inputs` layers before writing image
                 output_slices=None,         # list (of list) of slices to apply to `image_for_outputs` layers before writing image
                 output_target_shapes=None): # list of shapes of the target/gt images that correspond to all model outputs
        super(CARETensorBoard, self).__init__()
        is_tf_backend() or _raise(RuntimeError('TensorBoard callback only works with the TensorFlow backend.'))
        backend_channels_last() or _raise(NotImplementedError())
        IS_TF_1 or _raise(NotImplementedError("Not supported with TensorFlow 2"))

        self.freq = freq
        self.image_freq = freq
        self.prob_out = prob_out
        self.merged = None
        self.gt_outputs = None
        self.write_graph = write_graph
        self.write_images = write_images
        self.n_images = n_images
        self.image_for_inputs = image_for_inputs
        self.image_for_outputs = image_for_outputs
        self.input_slices = input_slices
        self.output_slices = output_slices
        self.output_target_shapes = output_target_shapes
        self.compute_histograms = compute_histograms

        if prefix_with_timestamp:
            log_dir = os.path.join(log_dir, datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S.%f"))

        self.log_dir = log_dir

    def set_model(self, model):
        self.model = model
        self.sess = K.get_session()
        tf_sums = []

        if self.compute_histograms and self.freq and self.merged is None:
            for layer in self.model.layers:
                for weight in layer.weights:
                    tf_sums.append(tf.summary.histogram(weight.name, weight))

                if hasattr(layer, 'output'):
                    tf_sums.append(tf.summary.histogram('{}_out'.format(layer.name),
                                                        layer.output))

        def _gt_shape(output_shape,target_shape):
            if target_shape is not None:
                output_shape = target_shape
            if not self.prob_out: return output_shape
            output_shape[-1] % 2 == 0 or _raise(ValueError())
            return list(output_shape[:-1]) + [output_shape[-1] // 2]

        n_inputs, n_outputs = len(self.model.inputs), len(self.model.outputs)
        image_for_inputs  = np.arange(n_inputs)  if self.image_for_inputs  is None else self.image_for_inputs
        image_for_outputs1 = np.arange(n_outputs) if self.image_for_outputs is None else self.image_for_outputs
        image_for_outputs2 = np.arange(n_outputs) if self.image_for_outputs is None else self.image_for_outputs

        output_target_shapes1 = [None]*n_outputs if self.output_target_shapes is None else self.output_target_shapes
        output_target_shapes2 = [None]*n_outputs if self.output_target_shapes is None else self.output_target_shapes
        self.gt_outputs1 = [K.placeholder(shape=_gt_shape(K.int_shape(x),sh)) for x,sh in zip(self.model.outputs,output_target_shapes1)]
        self.gt_outputs2 = [K.placeholder(shape=_gt_shape(K.int_shape(x),sh)) for x,sh in zip(self.model.outputs,output_target_shapes2)]

        input_slices  = (slice(None),) if self.input_slices  is None else self.input_slices
        output_slices1 = (slice(None),) if self.output_slices is None else self.output_slices
        output_slices2 = (slice(None),) if self.output_slices is None else self.output_slices
        if isinstance(input_slices[0],slice): # apply same slices to all inputs
            input_slices = [input_slices]*len(image_for_inputs)
        if isinstance(output_slices1[0],slice): # apply same slices to all outputs
            output_slices1 = [output_slices1]*len(image_for_outputs1)
            output_slices2 = [output_slices2]*len(image_for_outputs2)
        len(input_slices)  == len(image_for_inputs)  or _raise(ValueError())
        len(output_slices1) == len(output_slices2) == len(image_for_outputs1) == len(image_for_outputs2) or _raise(ValueError())

        def _name(prefix, layer, i, n, show_layer_names=False):
            return '{prefix}{i}{name}'.format (
                prefix = prefix,
                i      = (i if n > 1 else ''),
                name   = '' if (layer is None or not show_layer_names) else '_'+''.join(layer.name.split(':')[:-1]),
            )

        # inputs
        for i,sl in zip(image_for_inputs,input_slices):
            layer_name = _name('net_input', self.model.inputs[i], i, n_inputs)
            input_layer = tf_normalize_layer(self.model.inputs[i][tuple(sl)])
            tf_sums.append(tf.summary.image(layer_name, input_layer, max_outputs=self.n_images))

        # outputs
        for i1,i2,sl1,sl2 in zip(image_for_outputs1,image_for_outputs2,output_slices1,output_slices2):
            output_shape = self.model.output_shape if n_outputs==1 else self.model.output_shape[i1]
            # target
            output_layer1 = tf_normalize_layer(self.gt_outputs1[i1][tuple(sl1)])
            output_layer2 = tf_normalize_layer(self.gt_outputs2[i2][tuple(sl2)])
            layer_name1 = _name('net_target1', self.model.outputs[i1][0], i1, n_outputs)
            layer_name2 = _name('net_target2', self.model.outputs[i2][1], i2, n_outputs)
            tf_sums.append(tf.summary.image(layer_name1, output_layer1, max_outputs=self.n_images))
            tf_sums.append(tf.summary.image(layer_name2, output_layer2, max_outputs=self.n_images))
            # prediction
            n_channels_out = sep = output_shape[-1]
            if self.prob_out: # first half of output channels is mean, second half scale
                n_channels_out % 2 == 0 or _raise(ValueError())
                sep = sep // 2
            output_layer1 = tf_normalize_layer(self.model.outputs[i1][0][...,:sep][tuple(sl1)])
            output_layer2 = tf_normalize_layer(self.model.outputs[i2][1][...,:sep][tuple(sl2)])
            if self.prob_out:
                scale_layer1 = tf_normalize_layer(self.model.outputs[i1][0][...,sep:][tuple(sl1)], pmin=0, pmax=100)
                mean_name1  = _name('net_output_mean1',  self.model.outputs[i1[0]], i1, n_outputs)
                scale_name1 = _name('net_output_scale1', self.model.outputs[i1][0], i1, n_outputs)
                scale_layer2 = tf_normalize_layer(self.model.outputs[i2][1][...,sep:][tuple(sl2)], pmin=0, pmax=100)
                mean_name2  = _name('net_output_mean2',  self.model.outputs[i2[1]], i2, n_outputs)
                scale_name2 = _name('net_output_scale2', self.model.outputs[i2][1], i2, n_outputs)
                tf_sums.append(tf.summary.image(mean_name1, output_layer1, max_outputs=self.n_images))
                tf_sums.append(tf.summary.image(scale_name1, scale_layer1, max_outputs=self.n_images))
                tf_sums.append(tf.summary.image(mean_name2, output_layer2, max_outputs=self.n_images))
                tf_sums.append(tf.summary.image(scale_name2, scale_layer2, max_outputs=self.n_images))
            else:
                layer_name1 = _name('net_output1', self.model.outputs[i1][0], i1, n_outputs)
                layer_name2 = _name('net_output2', self.model.outputs[i2][0], i2, n_outputs)

                tf_sums.append(tf.summary.image(layer_name1, output_layer1, max_outputs=self.n_images))
                tf_sums.append(tf.summary.image(layer_name2, output_layer2, max_outputs=self.n_images))


        with tf.name_scope('merged'):
            self.merged = tf.summary.merge(tf_sums)

        with tf.name_scope('summary_writer'):
            if self.write_graph:
                self.writer = tf.summary.FileWriter(self.log_dir,
                                                    self.sess.graph)
            else:
                self.writer = tf.summary.FileWriter(self.log_dir)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}

        if self.validation_data and self.freq:
            if epoch % self.freq == 0:
                tensors = self.model.inputs + self.gt_outputs + self.model.sample_weights

                if self.model.uses_learning_phase:
                    tensors += [K.learning_phase()]
                    val_data = list(v[:self.n_images] for v in self.validation_data[:-2])
                    val_data += self.validation_data[-2:]
                else:
                    val_data = list(v[:self.n_images] for v in self.validation_data)

                feed_dict = dict(zip(tensors, val_data))
                result = self.sess.run([self.merged], feed_dict=feed_dict)
                summary_str = result[0]

                self.writer.add_summary(summary_str, epoch)

        for name, value in logs.items():
            if name in ['batch', 'size']:
                continue
            summary = tf.Summary()
            summary_value = summary.value.add()
            summary_value.simple_value = float(value)
            summary_value.tag = name
            self.writer.add_summary(summary, epoch)
        self.writer.flush()

    def on_train_end(self, _):
        self.writer.close()



class CARETensorBoardImage(Callback):

    def __init__(self, model, data, log_dir,
                 n_images=3,
                 batch_size=1,
                 prob_out=False,
                 image_for_inputs=None,  # write images for only these input indices
                 image_for_outputs=None, # write images for only these output indices
                 input_slices=None,      # list (of list) of slices to apply to `image_for_inputs` layers before writing image
                 output_slices=None):    # list (of list) of slices to apply to `image_for_outputs` layers before writing image

        super(CARETensorBoardImage, self).__init__()
        backend_channels_last() or _raise(NotImplementedError())
        not IS_TF_1 or _raise(NotImplementedError("Not supported with TensorFlow 1"))

        ## model
        from csbdeep.models.base_model import BaseModel
        if isinstance(model, BaseModel):
            model = model.keras_model
        isinstance(model, keras.Model) or _raise(ValueError())
        self.set_model(model) # changed here
        self.n_inputs  = len(self.model.inputs)
        self.n_outputs = len(self.model.outputs)

        if isinstance(data,(list,tuple)):
            X = data[0]['input']
            Y1 = itemgetter('dist1','prob1')(data[1])
            Y2 = itemgetter('dist2','prob2')(data[1])
        else:
            X = data[0]['input']
            Y1 = itemgetter('dist1','prob1')(data[1])
            Y2 = itemgetter('dist2','prob2')(data[1])
        if self.n_inputs  == 1 and isinstance(X,np.ndarray): X = (X,)
        if self.n_outputs == 1 and isinstance(Y1,np.ndarray) and isinstance(Y2,np.ndarray): 
            Y1 = (Y1,)
            Y2 = (Y2,)
        
        (self.n_inputs == len(X) and self.n_outputs/2 == len(Y1) == len(Y2)) or _raise(ValueError())
        self.X = [v[:n_images] for v in X]
        self.Y1 = [v[:n_images] for v in Y1]
        self.Y2 = [v[:n_images] for v in Y2]

        self.log_dir = log_dir
        self.batch_size = batch_size
        self.prob_out = prob_out

        ## I/O choices
        image_for_inputs  = np.arange(self.n_inputs)  if image_for_inputs  is None else image_for_inputs
        image_for_outputs1 = np.arange(self.n_outputs/2) if image_for_outputs is None else image_for_outputs
        image_for_outputs2 = np.arange(self.n_outputs/2) if image_for_outputs is None else image_for_outputs

        input_slices  = (slice(None),) if input_slices  is None else input_slices
        output_slices1 = (slice(None),) if output_slices is None else output_slices[:2]
        output_slices2 = (slice(None),) if output_slices is None else output_slices[2:]
        if isinstance(input_slices[0],slice): # apply same slices to all inputs
            input_slices = [input_slices]*len(image_for_inputs)
        if isinstance(output_slices1[0],slice): # apply same slices to all outputs
            output_slices1 = [output_slices1]*len(image_for_outputs1)
            output_slices2 = [output_slices2]*len(image_for_outputs2)

        len(input_slices)  == len(image_for_inputs)  or _raise(ValueError())
        len(output_slices1) == len(output_slices2) == len(image_for_outputs1) == len(image_for_outputs2) or _raise(ValueError())

        self.image_for_inputs = image_for_inputs
        self.image_for_outputs1 = image_for_outputs1
        self.image_for_outputs2 = image_for_outputs2
        self.input_slices = input_slices
        self.output_slices1 = output_slices1
        self.output_slices2 = output_slices2
        self.file_writer = tf.summary.create_file_writer(str(self.log_dir))


    def _name(self, prefix, layer, i, n, show_layer_names=False):
        return '{prefix}{i}{name}'.format (
            prefix = prefix,
            i      = (i if n > 1 else ''),
            name   = '' if (layer is None or not show_layer_names) else '_'+''.join(layer.name.split(':')[:-1]),
        )


    def _normalize_image(self, x, pmin=1, pmax=99.8, clip=True):
        def norm(x, axis):
            return normalize(x, pmin=pmin, pmax=pmax, axis=axis, clip=clip)

        n_channels_out = x.shape[-1]
        n_dim_out = len(x.shape)
        assert 0 < n_channels_out

        if n_dim_out > 4:
            x = np.max(x, axis=tuple(range(1,1+n_dim_out-4)))

        if n_channels_out == 1:
            out = norm(x, axis=(1,2))
        elif n_channels_out == 2:
            out = norm(np.concatenate([x,x[...,:1]], axis=-1), axis=(1,2,3))
        elif n_channels_out == 3:
            out = norm(x, axis=(1,2,3))
        else:
            out = norm(np.max(x, axis=-1, keepdims=True), axis=(1,2,3))
        return out


    def on_epoch_end(self, epoch, logs=None):
        # https://www.tensorflow.org/tensorboard/image_summaries

        # inputs
        if epoch == 0:
            with self.file_writer.as_default():
                for i,sl in zip(self.image_for_inputs,self.input_slices):
                    input_name = self._name('net_input', self.model.inputs[i], i, self.n_inputs)
                    input_image = self._normalize_image(self.X[i][tuple(sl)])
                    tf.summary.image(input_name, input_image, step=epoch)

        # outputs
        temp = self.model.predict(self.X, batch_size=self.batch_size)
        Y1hat = (temp['prob1'], temp['dist1'])
        Y2hat = (temp['prob2'], temp['dist2'])

        if self.n_outputs == 1 and isinstance(Y1hat,np.ndarray) and isinstance(Y2hat, np.ndarray): 
            Y1hat = (Y1hat,)
            Y2hat = (Y2hat,)
        with self.file_writer.as_default():
            for i1,i2,sl1,sl2 in zip(self.image_for_outputs1, self.image_for_outputs2,
            self.output_slices1, self.output_slices2):
                i1 = int(i1)
                i2 = int(i2)
                output_shape = self.model.output_shape if self.n_outputs==4 else self.model.output_shape[i1]
            
                n_channels_out = sep = len(output_shape)
                if self.prob_out: # first half of output channels is mean, second half scale
                    n_channels_out % 2 == 0 or _raise(ValueError())
                    sep = sep // 2
                output_image1 = self._normalize_image(Y1hat[i1][...,:sep][tuple(sl1)])
                output_image2 = self._normalize_image(Y2hat[i2][...,:sep][tuple(sl2)])
                if self.prob_out:
                    scale_image1 = self._normalize_image(Y1hat[i1][...,sep:][tuple(sl1)], pmin=0, pmax=100)
                    scale_name1 = self._name('net_output_scale1', self.model.outputs[i1][0], i1, self.n_outputs)
                    output_name1 = self._name('net_output_mean1',  self.model.outputs[i1][0], i1, self.n_outputs)
                    scale_image2 = self._normalize_image(Y2hat[i2][...,sep:][tuple(sl2)], pmin=0, pmax=100)
                    scale_name2 = self._name('net_output_scale2', self.model.outputs[i2][1], i2, self.n_outputs)
                    output_name2 = self._name('net_output_mean2',  self.model.outputs[i2][1], i2, self.n_outputs)
                    tf.summary.image(output_name1, output_image1, step=epoch)
                    tf.summary.image(scale_name1, scale_image1, step=epoch)
                    tf.summary.image(output_name2, output_image2, step=epoch)
                    tf.summary.image(scale_name2, scale_image2, step=epoch)
                else:
                    output_name1 = self._name('net_output1', self.model.outputs[i1][0], i1, self.n_outputs)
                    output_name2 = self._name('net_output2', self.model.outputs[i2][1], i2, self.n_outputs)
                    tf.summary.image(output_name1, output_image1, step=epoch)
                    tf.summary.image(output_name2, output_image2, step=epoch)

        # targets
        if epoch == 0:
            with self.file_writer.as_default():
                for i1,i2,sl1,sl2 in zip(self.image_for_outputs1,self.image_for_outputs2,
                self.output_slices1,self.output_slices2):
                    i1 = int(i1)
                    i2 = int(i2)
                    target_name1 = self._name('net_target1', self.model.outputs[i1][0], i1, self.n_outputs)
                    target_image1 = self._normalize_image(self.Y1[i1][tuple(sl1)])
                    target_name2 = self._name('net_target2', self.model.outputs[i2][1], i2, self.n_outputs)
                    target_image2 = self._normalize_image(self.Y2[i2][tuple(sl2)])
                    tf.summary.image(target_name1, target_image1, step=epoch)
                    tf.summary.image(target_name2, target_image2, step=epoch)


def write_prediction_summary(save_path, well_diameter_mm, px_per_mm, counts_per_frame, avg_areas_per_frame):
    """
    Write the prediction summary to a text file.

    Format includes header, settings (diameter, scale), and a per-frame table
    of plaque counts and average areas for each well.

    Args:
        save_path (str): File path to save the summary to.
        well_diameter_mm (float): Assumed well diameter in millimeters.
        px_per_mm (float): Calculated pixel-to-millimeter scale factor.
        counts_per_frame (list): List of lists, where each inner list contains plaque counts for the 6 wells.
        avg_areas_per_frame (list): List of lists, where each inner list contains average plaque areas (in mm^2).

    Returns:
        None
    """
    with open(save_path, "w") as f:
        f.write(f"PREDICTION SUMMARY\n")
        f.write("="*40 + "\n")
        f.write(f"Assumed Well Diameter: {well_diameter_mm} mm\n")
        f.write(f"Pixel-to-mm Scale: {px_per_mm:.2f} px/mm\n")

        for frame_idx, (counts, avg_areas) in enumerate(zip(counts_per_frame, avg_areas_per_frame)):
            f.write(f"\nFRAME {frame_idx + 1}\n")
            f.write("Plaque Count per Well      Average Plaque Area (mm²)\n")
            f.write("-------------------        ----------------------------\n")
            f.write("| {:03d} | {:03d} | {:03d} |        | {:06.2f} | {:06.2f} | {:06.2f} |\n".format(*counts[:3], *avg_areas[:3]))
            f.write("-------------------        ----------------------------\n")
            f.write("| {:03d} | {:03d} | {:03d} |        | {:06.2f} | {:06.2f} | {:06.2f} |\n".format(*counts[3:], *avg_areas[3:]))
            f.write("-------------------        ----------------------------\n")



from abc import ABC, abstractmethod

import numpy as np
import tensorflow as tf
from scipy.spatial.distance import pdist
from tensorflow.keras import backend as K
from tensorflow.keras.metrics import (
    BinaryCrossentropy, CategoricalAccuracy, CategoricalCrossentropy, MeanSquaredError, Metric
)


def get_metric(
    metric_name: str,
    **kwargs,
    ) -> Metric:
    if metric_name.lower() == "dice" | "dicecoeff" | "dicecoefficient":
            return DiceCoeff(**kwargs)
    if metric_name.lower() == "mse":
            return MeanSquaredError(**kwargs)
    if metric_name.lower() == "cce" | "categoricalcrossentropy":
            return CategoricalCrossentropy(**kwargs)
    if metric_name.lower() == "bce" | "binarycrossentropy":
            return BinaryCrossentropy(**kwargs)
    if metric_name.lower() == "categoricalaccuracy":
            return CategoricalAccuracy(**kwargs)
    if metric_name.lower() == _:
            raise ValueError(f"Unknown metric: {metric_name}")


def dice_coeff(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    smooth: float = 1.,
    ) -> tf.Tensor:
    """ Calculate dice coefficient.

        For binary segmentation n_classes should be equal to 1, and not ommited.
    Args:
        y_true (tf.Tensor): Ground truth values.    shape = [batch_size, W, H, n_classes]
        y_pred (tf.Tensor): Predicted values.       shape = [batch_size, W, H, n_classes]
        smooth (float): Smoothing factor to avoid division by zero.
    Returns:
        tf.Tensor: Dice coefficient value.          shape = [n_classes]
    """

    axes = [0, 1, 2]

    intersections = K.sum(y_pred * y_true, axis=axes)
    scores = (2. * intersections + smooth) / (K.sum(y_true, axis=axes) + K.sum(y_pred, axis=axes) + smooth)

    return scores

class MeanMetric(Metric, ABC):
    def __init__(self, name=None, **kwargs):
        super().__init__(name=name, **kwargs)
        self.scores = []

    def reset_state(self):
        self.scores = []

    @abstractmethod
    def _calculate_metric(self, *args):
        raise NotImplementedError()

    def update_state(self, *args, sample_weight=None):
        metric = self._calculate_metric(*args)
        if sample_weight is not None:
            metric *= sample_weight
        self.scores.append(metric)

    def result(self):
        return tf.math.reduce_mean(self.scores)

class Alignment(MeanMetric):
    """ Calculates alignment between an image and its augmentation.
    """
    def __init__(self, name="alignment", alpha=2, **kwargs):
        super().__init__(name=name, **kwargs)
        self.alpha = alpha

    def _calculate_metric(self, x, y):
        return np.mean(np.power(np.linalg.norm(x-y, ord=2, axis=1), self.alpha))

class Uniformity(MeanMetric):
    """ Calculates whether the embeddings are uniformly distributed on the unit hypersphere.
    """
    def __init__(self, name="uniformity", t=2, **kwargs):
        super().__init__(name=name, **kwargs)
        self.t = t

    def _calculate_metric(self, x):
        sq_pdist = pdist(x.reshape(x.shape[0], -1), "euclidean")**2
        return np.log(np.exp(-self.t * sq_pdist).mean(axis=0))

class DiceCoeff(MeanMetric):
    def __init__(self, name="dice_coeff", **kwargs):
        super().__init__(name=name, **kwargs)

    def _calculate_metric(self, y_true, y_pred):
        return dice_coeff(y_true, y_pred)
